# -*- coding: utf-8 -*-
"""
Compiled queries ready for execution.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from weakref import WeakKeyDictionary

from relstorage._util import CachedIn

from ._util import copy
from ._util import Columns
from .dialect import DialectAware

from .expressions import ParamMixin
from .expressions import And
from .expressions import EmptyExpression

class Clause(DialectAware):
    """
    A portion of a SQL statement.
    """

class ColumnList(Columns):
    """
    List of columns used in a query.
    """

    # This class exists for semantics, it currently doesn't
    # do anything different than the super.

class WhereClause(Clause):

    def __init__(self, expression):
        self.expression = expression

    def and_(self, expression):
        expression = And(self.expression, expression)
        new = copy(self)
        new.expression = expression
        return new

    def __compile_visit__(self, compiler):
        compiler.emit_keyword('WHERE')
        compiler.visit_grouped(self.expression)


class OrderBy(Clause):

    def __init__(self, expression, dir):
        self.expression = expression
        self.dir = dir

    def __compile_visit__(self, compiler):
        compiler.emit_keyword('ORDER BY')
        compiler.visit(self.expression)
        if self.dir:
            compiler.emit(' ' + self.dir)


def where(expression):
    if expression:
        return WhereClause(expression)

class WhereMixin(object):
    _where = EmptyExpression()

    def where(self, expression):
        expression = expression.resolve_against(self.table)
        s = copy(self)
        s._where = where(expression)
        return s

    def and_(self, expression):
        expression = expression.resolve_against(self.table)
        s = copy(self)
        s._where = self._where.and_(expression)
        return s

class Query(ParamMixin,
            Clause):
    __name__ = None
    prepare = False

    def __str__(self):
        return str(self.compiled())

    def __get__(self, inst, klass):
        if inst is None:
            return self
        # We need to set this into the instance's dictionary.
        # Otherwise we'll be rebinding and recompiling each time we're
        # accessed which is not good (in fact it's a step backwards
        # from query_property()). On Python 3.6+, there's the
        # `__set_name__(klass, name)` called on the descriptor which
        # does the job perfectly. In earlier versions, we're on our
        # own.
        #
        # TODO: In test cases, we spend a lot of time binding and
        # compiling. Can we find another layer of caching somewhere?
        # The dialect object would probably work.

        if not self.__name__:
            # Go through the class hierarchy, find out what we're called.
            for base in klass.mro():
                for k, v in vars(base).items():
                    if v is self:
                        self.__name__ = k
                        break
        assert self.__name__

        result = self.bind(inst).compiled()
        vars(inst)[self.__name__] = result

        return result

    def __set_name__(self, owner, name):
        self.__name__ = name

    @CachedIn('_v_compiled')
    def compiled(self):
        return CompiledQuery(self)

    def prepared(self):
        """
        Note that it's good to prepare this query, if
        supported by the driver.
        """
        s = copy(self)
        s.prepare = True
        return s

    def execute(self, cursor):
        raise AttributeError("Do not execute a Query without compiling it.")


class CompiledQuery(object):
    """
    Represents a completed query.
    """

    stmt = None
    params = None
    _raw_stmt = None
    _prepare_stmt = None
    _prepare_converter = None

    def __init__(self, root):
        self.root = root
        # We do not keep a reference to the context;
        # it's likely to be an instance object that's
        # going to have us stored in its dictionary.
        dialect = root.dialect

        compiler = dialect.compiler(root)
        self.stmt, self.params = compiler.compile()
        self._raw_stmt = self.stmt # for debugging
        if compiler.can_prepare():
            self._prepare_stmt, self.stmt, self._prepare_converter = compiler.prepare()

    def __repr__(self):
        if self._prepare_stmt:
            return "%s (%s)" % (
                self.stmt,
                self._prepare_stmt
            )
        return self.stmt

    def __str__(self):
        return self.stmt

    _connection_cache = WeakKeyDictionary()

    def _stmt_cache_for_connection(self, connection):
        """Returns a dictionary."""
        # If we can't store it directly on the cursor, as happens for
        # types implemented in C, we use a weakkey dictionary.
        try:
            session_prep_stmts = connection._rs_prepared_statements
        except AttributeError:
            try:
                session_prep_stmts = connection._rs_prepared_statements = {}
            except AttributeError:
                session_prep_stmts = self._connection_cache.get(connection)
                if session_prep_stmts is None:
                    session_prep_stmts = self._connection_cache[connection] = {}
        return session_prep_stmts

    def execute(self, cursor, params=None):
        # (Any, dict) -> None
        # TODO: Include literals from self.params.
        # TODO: Syntax transformation if they don't support names.
        # TODO: Validate given params match expected ones, nothing missing?
        stmt = self.stmt
        if self._prepare_stmt:
            # Prepare on demand.

            # In all databases, prepared statements
            # persist past COMMIT/ROLLBACK (though in PostgreSQL
            # preparing them takes locks that only go away at
            # COMMIT/ROLLBACK). But they don't persist past a session
            # restart (new connection) (obviously).
            #
            # Thus we keep a cache of statements we have prepared for
            # this particular connection/cursor.
            # TODO: This should probably really be on the connection,
            # not the cursor. connection is session.
            session_prep_stmts = self._stmt_cache_for_connection(cursor.connection)
            try:
                stmt = session_prep_stmts[self._prepare_stmt]
            except KeyError:
                stmt = session_prep_stmts[self._prepare_stmt] = self.stmt
                __traceback_info__ = self._prepare_stmt, self, self.root.dialect.compiler(self.root)
                cursor.execute(self._prepare_stmt)
            params = self._prepare_converter(params)

        if params:
            cursor.execute(stmt, params)
        elif self.params:
            # XXX: This isn't really good.
            # If there are both literals in the SQL and params,
            # we don't handle that.
            cursor.execute(stmt, self.params)
        else:
            cursor.execute(stmt)

    def executemany(self, cursor, paramlist):
        return cursor.executemany(self.stmt, paramlist)



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
IOIDAllocator implementations.
"""

from __future__ import absolute_import

from zope.interface import implementer

from ..._compat import metricmethod
from ..interfaces import IOIDAllocator
from ..oidallocator import AbstractRangedOIDAllocator


@implementer(IOIDAllocator)
class OracleOIDAllocator(AbstractRangedOIDAllocator):

    def __init__(self, connmanager):
        self.connmanager = connmanager

    def _set_min_oid_from_range(self, cursor, n):
        """Ensure the next OID is at least the given OID."""
        stmt = "SELECT zoid_seq.nextval FROM DUAL"
        cursor.execute(stmt)
        next_n = cursor.fetchone()[0]
        if next_n < n:
            # Oracle provides no way modify the sequence value
            # except through alter sequence or drop/create sequence,
            # but either statement kills the current transaction.
            # Therefore, open a temporary connection to make the
            # alteration.
            conn2, cursor2 = self.connmanager.open()
            try:
                # Change the sequence by altering the increment.
                # (this is safer than dropping and re-creating the sequence)
                diff = n - next_n
                cursor2.execute(
                    "ALTER SEQUENCE zoid_seq INCREMENT BY %d" % diff)
                cursor2.execute("SELECT zoid_seq.nextval FROM DUAL")
                cursor2.execute("ALTER SEQUENCE zoid_seq INCREMENT BY 1")
                conn2.commit()
            finally:
                self.connmanager.close(conn2, cursor2)

    @metricmethod
    def new_oids(self, cursor):
        """Return a sequence of new, unused OIDs."""
        stmt = "SELECT zoid_seq.nextval FROM DUAL"
        cursor.execute(stmt)
        n = cursor.fetchone()[0]
        return self._oid_range_around(n)

    def reset_oid(self, cursor):
        raise NotImplementedError



# -*- coding: utf-8 -*-
"""
Syntax elements.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from ._util import Resolvable


class LiteralNode(Resolvable):

    __slots__ = (
        'raw',
        'name',
    )

    def __init__(self, raw):
        self.raw = raw
        self.name = 'anon_%x' % (id(self),)

    @property
    def alias(self):
        return self.name

    def __compile_visit__(self, compiler):
        compiler.emit(str(self.raw))

    def resolve_against(self, table):
        return self

class NullNode(LiteralNode):
    __slots__ = ()

    def __init__(self):
        super(NullNode, self).__init__(None)

    def __compile_visit__(self, compiler):
        compiler.emit_null()

class BooleanNode(LiteralNode):

    __slots__ = ()

class TextNode(LiteralNode):
    __slots__ = ()

def as_node(c):
    if c is None:
        return NullNode()
    if isinstance(c, bool):
        return BooleanNode(c)
    if isinstance(c, int):
        return LiteralNode(c)
    if isinstance(c, str):
        return TextNode(c)
    return c



def resolved_against(columns, table):
    resolved = [
        as_node(c).resolve_against(table)
        for c
        in columns
    ]
    return resolved



##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Test mixin for exercising persistent caching features.
"""
import transaction

from persistent.mapping import PersistentMapping
from ZODB.DB import DB
from ZODB.serialize import referencesf
#from ZODB.utils import p64 as int64_to_8bytes
from ZODB.utils import u64 as bytes8_to_int64
from ZODB.utils import z64

from . import TestCase


ROOT_OID = 0
ROOT_KEY = 'myobj'

def find_cache(obj):
    # Pass a connection, a storage, or the cache.
    storage = getattr(obj, '_storage', obj)
    cache = getattr(storage, '_cache', storage)
    return cache

def as_int(oid):
    if isinstance(oid, bytes):
        oid = bytes8_to_int64(oid)
    return oid

class PersistentCacheStorageTests(TestCase):
    # pylint:disable=abstract-method

    _storage = None

    def make_storage(self, *args, **kwargs):
        raise NotImplementedError

    def __make_storage_pcache(
            self,
            expected_checkpoints=None,
            expected_root_tid=None,
    ): # pylint:disable=unused-argument
        """
        Make a storage that reads a persistent cache that
        already exists.
        """
        storage = self.make_storage(zap=False)
        cache = storage._cache
        # It did read checkpoints and TID
        # self.assertIsNotNone(cache.checkpoints)
        # if expected_checkpoints:
        #     self.assertEqual(cache.checkpoints, expected_checkpoints)

        if expected_root_tid is not None:
            # Nothing is actually in the index
            self.assert_oid_not_known(ROOT_OID, storage)
            # because everything is frozen.
            _state, tid = cache.local_client[(ROOT_OID, None)]
            self.assertEqual(tid, expected_root_tid)
        else:
            self.assert_oid_not_known(ROOT_OID, storage)
        self.assertIsNotNone(cache.polling_state.object_index.maximum_highest_visible_tid)
        return storage

    def __make_storage_no_pcache(self):
        storage = self._closing(
            self.make_storage(cache_local_dir=None, zap=False)
        )
        # It didn't read checkpoints or TID
        cache = find_cache(storage)
        # self.assertIsNone(cache.checkpoints)
        self.assertIsNone(cache.current_tid)
        return storage

    def assert_oid_not_known(self, oid, storage):
        cache = find_cache(storage)
        self.assertNotIn(oid, cache.object_index or ())

    def assert_exact_tid_in_oid_index(self, oid, storage, expected_tid=None):
        oid = as_int(oid)
        expected_tid = as_int(expected_tid)
        cache = find_cache(storage)
        index = cache.object_index or cache.polling_state.object_index or ()
        __traceback_info__ = index.keys()
        self.assertIn(oid, index)
        tid = index[oid]
        if expected_tid:
            self.assertEqual(tid, expected_tid)
        return tid

    assert_oid_known = assert_exact_tid_in_oid_index

    def assert_tid_after(self, oid, tid, storage):
        cache = find_cache(storage)
        new_tid = self.assert_oid_known(oid, cache)
        self.assertGreater(new_tid, tid)
        return new_tid

    def assert_oid_current(self, oid, storage, expected_tid=None):
        """
        Check that *oid* is in the object index; if *expected_tid* is given,
        make sure it matches that value.
        """
        cache = find_cache(storage)

        tid = self.assert_oid_known(oid, cache)
        self.assertEqual(cache.current_tid, tid,
                         "cache tid is not OID tid (expected_tid: %s)" % (expected_tid,))
        if expected_tid is not None:
            self.assertEqual(tid, expected_tid)
        return tid

    # def assert_checkpoints(self, storage, expected_checkpoints=None):
    #     cache = find_cache(storage)
    #     self.assertIsNotNone(cache.checkpoints)
    #     if expected_checkpoints:
    #         self.assertEqual(cache.checkpoints, expected_checkpoints)
    #     return cache.checkpoints

    def assert_cached(self, oid, tid, storage, msg=None):
        oid = as_int(oid)
        tid = as_int(tid)
        cache = find_cache(storage)
        cache_data = cache.local_client[(oid, tid)]
        __traceback_info__ = (
            (oid, tid),
            cache.object_index,
            cache.polling_state.object_index,
            [k
             for k in cache.local_client
             if k[0] == oid]
        )
        self.assertIsNotNone(cache_data,
                             msg or "Unable to find cache for (%s, %s)" % (oid, tid))
        return cache_data

    def assert_cached_exact(self, oid, tid, storage, msg=None):
        data = self.assert_cached(oid, tid, storage, msg)
        self.assertEqual(data[1], as_int(tid), msg)
        return data

    def assert_oid_not_cached(self, oid, storage, msg=None):
        oid = as_int(oid)
        cache = find_cache(storage)
        keys = [k for k in cache.local_client if k[0] == oid]
        __traceback_info__ = (oid, keys)
        self.assertEmpty(keys, msg)


    def __do_sets(self, root, new_data, old_tids, old_data):
        for key, value in new_data.items():
            old_tid = old_tids.get(key)
            old_value = old_data.get(key)

            key_path = key.split('.')
            attr_name = key_path[-1]
            __traceback_info__ = key_path
            base = root
            name = None
            for name in key_path[:-1]:
                base = getattr(base, name)
            oid = bytes8_to_int64(base._p_oid)
            __traceback_info__ = key, oid, old_tid
            if old_tid is not None:
                # Opening the database loaded the root object, so it's
                # now cached with the expected key; it may not actually
                # be at that exact TID, though.
                self.assert_cached(oid, old_tid, root._p_jar)

            if old_value is not None:
                val = getattr(base, attr_name)
                self.assertEqual(val, old_value)

            setattr(base, attr_name, value)
            # Make sure we have something
            old_tids[key] = old_tid

    def __do_check_tids(self, root, old_tids):
        for key, old_tid in old_tids.items():
            key_path = key.split('.')
            base = root
            for name in key_path[:-1]:
                base = getattr(base, name)
            oid = bytes8_to_int64(base._p_oid)

            # We have a saved TID for the root object. If we had an old one,
            # it's now bigger.
            if old_tid is not None:
                self.assert_tid_after(oid, old_tid, root._p_jar)
            else:
                self.assert_oid_current(oid, root._p_jar)

    def __set_keys_in_root_to(self,
                              storage,
                              new_data,
                              old_tids,
                              old_data,
                              pack=False):
        """
        Set the values for *new_data* in the root object of the storage.

        And return the transaction ID of when we mad the change,
        and the transaction ID of the last time the root changed.

        Uses an independent transaction.

        *old_tids* is a map from the keys in *new_data* to an expected TID
        that should be cached. *old_value* is the same for the expected
        current values on the root.

        Closes *storage*.
        """
        db1 = self._closing(DB(storage))
        tx = transaction.TransactionManager()
        c1 = db1.open(tx)
        # We've polled and gained checkpoints
        # self.assert_checkpoints(c1)

        root = c1.root()
        self.__do_sets(root, new_data, old_tids, old_data)
        tx.commit()

        self.__do_check_tids(root, old_tids)
        tid_int = bytes8_to_int64(c1._storage.lastTransaction())
        self.assertEqual(c1._storage._cache.current_tid, tid_int)
        c1.close()
        if pack:
            storage.pack(tid_int, referencesf)
        db1.close()
        return tid_int, bytes8_to_int64(root._p_serial)

    def __set_key_in_root_to(self,
                             storage,
                             value,
                             key=ROOT_KEY,
                             old_tid=None,
                             old_value=None):
        return self.__set_keys_in_root_to(
            storage,
            {key: value},
            {key: old_tid},
            {key: old_value}
        )

    def checkNoConflictWhenChangeMissedByPersistentCacheAfterCP0(self):
        orig_tid, orig_checkpoints = self.__set_key_in_root_to(self._storage, 42)

        # A storage that will not update the persistent cache.
        new_tid, _ = self.__set_key_in_root_to(
            self.__make_storage_no_pcache(),
            420
        )
        self.assertGreater(new_tid, orig_tid)

        # Now a new storage that will read the persistent cache.
        # It has the correct checkpoints,
        storage3 = self.__make_storage_pcache(
            orig_checkpoints)

        db3 = self._closing(DB(storage3))
        c3 = db3.open()

        # Polling in the connection, however, caught us up, because the object changed
        # after current_tid.
        self.assert_oid_current(ROOT_OID, c3, new_tid)

        r = c3.root()
        self.assertEqual(r.myobj, 420)
        r.myobj = 180
        transaction.commit()
        c3.close()
        db3.close()

    def _populate_root_and_mapping(self):
        """
        Creates the following structure in ``self._storage``::

            root.myobj1 = PersistentMapping()
            root.myobj1.key = PersistentMapping()
            root.myobj = 3

        Does this over several transactions. Returns
        the tid of the last time the root changed, and the tid
        of ``root.myobj1``, which is later than the root TID and which
        is current, and the database opened on the storage.
        """
        tx1 = transaction.TransactionManager()
        storage1 = self._storage
        db1 = self._closing(DB(storage1))
        c1 = db1.open(tx1)
        root = c1.root
        root().myobj1 = root.myobj1 = mapping = PersistentMapping()
        root().myobj = root.myobj = 1
        tx1.commit()
        c1._storage._cache.clear(load_persistent=False)

        c1._storage.poll_invalidations()
        root().myobj = root.myobj = 2
        tx1.commit()
        c1._storage._cache.clear(load_persistent=False)

        c1._storage.poll_invalidations()
        root().myobj = root.myobj = 3
        tx1.commit()
        root_tid = self.assert_oid_known(ROOT_OID, c1)
        c1._storage._cache.clear(load_persistent=False)

        # Now, mutate an object that's not the root
        # so that we get a new transaction after the root was
        # modified. This transaction will be included in
        # a persistent cache.
        c1._storage.poll_invalidations()
        root().myobj1.key = root.myobj1.key = PersistentMapping()
        mapping_oid = mapping._p_oid
        mapping_oid_int = bytes8_to_int64(mapping_oid)
        tx1.commit()
        mapping_tid = self.assert_oid_known(mapping_oid_int, c1)

        # self.assert_checkpoints(c1, (root_tid, root_tid))
        self.assert_oid_current(mapping_oid_int, c1)

        # the root is not in a delta
        self.assert_oid_not_known(ROOT_OID, c1)
        # Nor is it in the cache, because the Connection's
        # object cache still had the root and we were never
        # asked.
        self.assert_oid_not_cached(ROOT_OID, c1)
        # So lets get it in the cache with its current TID.
        c1._storage.load(z64)
        self.assert_cached_exact(ROOT_OID, root_tid, c1)

        c1.close()
        return root_tid, mapping_tid, db1

    def checkNoConflictWhenChangeMissedByPersistentCacheBeforeCP1(self):
        _root_tid, _mapping_tid, db = self._populate_root_and_mapping()

        # Make some changes to the root in a storage that will not
        # read or update the persistent cache.
        new_tid, _ = self.__set_key_in_root_to(
            self.__make_storage_no_pcache(),
            420,
            old_tid=None,
        )

        # Persist
        db.close()

        # Now a new storage that will read the persistent cache
        storage3 = self.__make_storage_pcache(
            expected_checkpoints=(new_tid, new_tid),
            expected_root_tid=None,
        )
        # The root object is not cached, it was rejected.
        self.assert_oid_not_cached(ROOT_OID, storage3)

        # We can successfully open and edit the root object.
        self.__set_key_in_root_to(storage3, 180, old_tid=new_tid, old_value=420)

    def checkNoConflictNoChangeInPersistentCacheBeforeCP1(self):
        root_tid, _mapping_tid, db = self._populate_root_and_mapping()

        # Make some changes to the sub-object, and not the root, in a storage that will not
        # read or update the persistent cache.
        new_tid, root_tid = self.__set_key_in_root_to(
            self.__make_storage_no_pcache(),
            420,
            key='myobj1.key',
        )

        # persist
        db.close()

        # Now a new storage that will read the persistent cache
        storage = self.__make_storage_pcache(
            # With checkpoints
            expected_checkpoints=(new_tid, new_tid),
            # No root TID in delta after 0
            expected_root_tid=None
        )
        # But it is in the cache for its correct key, because we verified it
        # to still be in sync.
        cache_data = self.assert_cached(ROOT_OID, root_tid, storage)
        self.assertEqual(cache_data[1], root_tid)

        self.__set_keys_in_root_to(
            storage,
            {
                'myobj': 180,
                'myobj1.key': 360,
            },
            old_data={
                'myobj': 3,
                'myobj1.key': 420
            },
            old_tids={}
        )

    def checkNoConflictWhenDeletedNotInInPersistentCacheBeforeCP1(self):
        root_tid, _mapping_tid, db = self._populate_root_and_mapping()

        # Now, remove a persistent object. We do this by setting its
        # key to a new persistent object.
        c1 = db.open()
        root = c1.root()
        new_nested_mapping = PersistentMapping()
        root.myobj1.key = new_nested_mapping

        mapping_oid = root.myobj1._p_oid
        mapping_oid_int = bytes8_to_int64(mapping_oid)
        c1.add(new_nested_mapping)
        nested_mapping_oid = new_nested_mapping._p_oid
        nested_mapping_oid_int = bytes8_to_int64(nested_mapping_oid)
        transaction.commit()
        self.assert_oid_current(nested_mapping_oid_int, c1)

        # self.assert_checkpoints(c1, (root_tid, root_tid))

        # the root is not in a delta
        # self.assert_oid_not_known(ROOT_OID, c1)

        # Though it is in the cache.
        self.assert_cached_exact(ROOT_OID, root_tid, c1)

        # Create a new transaction that derefs an object but
        # that won't update the persistent cache.
        new_tid, _ = self.__set_keys_in_root_to(
            self.__make_storage_no_pcache(),
            {'myobj1.key': None},
            {},
            {},
            pack=True
        )

        # Persist
        c1.close()
        db.close()
        del db, c1

        # Now a new storage that will read the persistent cache
        storage = self.__make_storage_pcache(
            expected_checkpoints=(new_tid, new_tid),
        )
        # The deleted object was not put in a delta map
        self.assert_oid_not_known(nested_mapping_oid_int, storage)
        # Though it remains cached.
        # self.assert_oid_not_cached(nested_mapping_oid_int, storage)

        # Likewise, the parent mapping isn't found anywhere, because it
        # changed
        self.assert_oid_not_known(mapping_oid_int, storage)
        # self.assert_oid_not_cached(mapping_oid_int, storage)

        self.__set_keys_in_root_to(
            storage,
            {'myobj': 180, 'myobj1.key': 360},
            {'': root_tid},
            {'myobj': 3, 'myobj1.key': None}
        )


    def checkNoConflictWhenOverlappedModificationNotInCache(self):
        # A storage writes a current state of an object O1 at TID1
        # to persistent cache.
        # Meanwhile, an unconnected storage writes O1
        # to the database at TID2.
        # A third storage polls, and sees the change for O1, but has
        # never had O1 in its cache. It writes the persistent cache.

        # When we open and read the persistent cache, we don't use the old
        # state for O1.

        # First storage caches this state.
        root_tid, _mapping_tid, db = self._populate_root_and_mapping()
        conn = db.open()
        nested_tid = conn.root.myobj1.key._p_serial
        nested_oid = conn.root.myobj1.key._p_oid
        conn.close()
        # Persist.
        db.close()

        # Second storage, disconnected, mutates.
        # Third storage is around to observe this.
        storage2 = self.__make_storage_no_pcache()
        txm2 = transaction.TransactionManager(True)
        db2 = DB(storage2)

        storage3 = self.__make_storage_pcache()
        find_cache(storage3).clear(load_persistent=False)
        storage3.poll_invalidations()

        conn2 = db2.open(txm2)
        txm2.begin()
        conn2.root.myobj1.key['foo'] = 'bar'

        # A fourth storage is opened right now, loading the old state.
        storage_with_old_state = self.__make_storage_pcache()

        txm2.commit()
        _, new_nested_tid = conn2._storage.load(nested_oid)
        self.assertNotEqual(nested_tid, new_nested_tid)
        conn2.close()
        db2.close()


        # Storage 3 observes the change, but has never had that object cached.
        storage3.poll_invalidations()
        self.assert_exact_tid_in_oid_index(nested_oid, storage3, new_nested_tid)
        self.assert_oid_not_cached(nested_oid, storage3)
        # Persist. Make it have a hit first so that it actually does want to save.
        storage3.load(z64)
        self.assertEqual(1, find_cache(storage3).save())
        storage3.close()

        storage4 = self.__make_storage_pcache()
        # When storage3 saved, it removed that data from the cache, so
        # storage4 doesn't have it.
        self.assert_oid_not_cached(
            nested_oid, storage4,
            "OID %s should not be cached except with TID %s" % (
                bytes8_to_int64(nested_oid),
                bytes8_to_int64(new_nested_tid)
            ))
        self.assert_cached(ROOT_OID, root_tid, storage4)

        def _check_nested(storage):
            db4 = DB(storage)
            txm4 = transaction.TransactionManager(True)
            conn4 = db4.open(txm4)
            txm4.begin()
            self.assertEqual(
                conn4.root.myobj1.key['foo'],
                'bar'
            )
            conn4.close()
            db4.close()
        _check_nested(storage4)

        # The one opened in between has it, but won't use it.
        self.assert_cached_exact(nested_oid, nested_tid, storage_with_old_state)

        # However, when it polls, it won't use the old state.
        _check_nested(storage_with_old_state)

    def checkNoConflictWhenOverlappedModificationNotInCache2(self):
        # pylint:disable=too-many-statements
        # Like checkNoConflictWhenOverlappedModificationNotInCache,
        # but the storage that does the invalidation polls
        # several times and loses the invalidation data from its object index.

        # First storage caches this state.
        root_tid, _mapping_tid, db = self._populate_root_and_mapping()
        conn = db.open()
        nested_tid = conn.root.myobj1.key._p_serial
        nested_oid = conn.root.myobj1.key._p_oid
        conn.close()
        # Persist.
        db.close()

        # Second storage, disconnected, mutates.
        # Third storage is around to observe this.
        storage2 = self.__make_storage_no_pcache()
        txm2 = transaction.TransactionManager(True)
        db2 = DB(storage2)

        storage3 = self.__make_storage_pcache()
        find_cache(storage3).clear(load_persistent=False)
        storage3.poll_invalidations()

        conn2 = db2.open(txm2)
        txm2.begin()
        conn2.root.myobj1.key['foo'] = 'bar'

        # A fourth storage is opened right now, loading the old state.
        storage_with_old_state = self.__make_storage_pcache()

        txm2.commit()
        _, new_nested_tid = conn2._storage.load(nested_oid)
        self.assertNotEqual(nested_tid, new_nested_tid)
        conn2.close()
        db2.close()


        # Storage 3 observes the change, but has never had that object cached.
        storage3.poll_invalidations()
        self.assert_exact_tid_in_oid_index(nested_oid, storage3, new_nested_tid)
        self.assert_oid_not_cached(nested_oid, storage3)

        # Storage 3 makes some changes of its own and polls forward, losing
        # the old object index.
        def _make_change(storage):
            db = DB(storage)
            tx = transaction.TransactionManager(True)
            conn = db.open(tx)
            tx.begin()
            conn.root.another_obj = 42
            tx.commit()
            tid = conn.root()._p_serial
            tx.begin()
            tx.commit()
            conn.close()
            return tid, db

        root_tid, db = _make_change(storage3)
        storage3.poll_invalidations()
        # We still have this data in our map because the DB object and its open
        # connection still have that around.
        self.assert_exact_tid_in_oid_index(nested_oid, storage3, new_nested_tid)
        self.assert_oid_not_cached(nested_oid, storage3)


        # Persist. Make it have a hit first so that it actually does want to save.
        storage3.load(z64)
        self.assertEqual(1, find_cache(storage3).save())
        # We detached everything and vacuumed in order to save. So
        # we've lost the tid data, and couldn't invalidate it based on
        # the object_index. So unless we do validation based on
        # something else, we'd still save and use incorrect data.
        self.assert_oid_not_known(nested_oid, storage3)
        storage3.close()

        storage4 = self.__make_storage_pcache()
        # We still figured out to drop this data, though.
        self.assert_oid_not_cached(
            nested_oid, storage4,
            "OID %s should not be cached except with TID %s" % (
                bytes8_to_int64(nested_oid),
                bytes8_to_int64(new_nested_tid)
            ))
        self.assert_cached(ROOT_OID, root_tid, storage4)

        def _check_nested(storage):
            db4 = DB(storage)
            txm4 = transaction.TransactionManager(True)
            conn4 = db4.open(txm4)
            txm4.begin()
            self.assertEqual(
                conn4.root.myobj1.key['foo'],
                'bar'
            )
            conn4.close()
            db4.close()
        _check_nested(storage4)

        # The one opened in between has it, but won't use it.
        self.assert_cached_exact(nested_oid, nested_tid, storage_with_old_state)

        # However, when it polls, it won't use the old state.
        _check_nested(storage_with_old_state)



# -*- coding: utf-8 -*-
"""
Interfaces, mostly internal, for the sql module.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint:disable=inherit-non-class,no-method-argument

from zope.interface import Interface

class ITypedParams(Interface):
    """
    Something that accepts parameters, and knows
    what their types should be.
    """

    def datatypes_for_parameters():
        """
        Returns a sequence of datatypes.

        XXX: This only works for ordered params; make this work
        for named parameters. Probably want to treat the two the same,
        with ordered parameters using indexes as their name.
        """


class IBindParam(Interface):
    """
    A parameter to a query.
    """

class INamedBindParam(IBindParam):
    """
    A named parameter.
    """

class IOrderedBindParam(IBindParam):
    """
    A anonymous parameter, identified only by order.
    """



##############################################################################
#
# Copyright (c) 2016 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from zope.interface import Attribute
from zope.interface import Interface


from transaction.interfaces import TransientError
from ZODB.POSException import StorageError

from relstorage.interfaces import IDetachableMVCCDatabaseViewer
from relstorage.interfaces import IMVCCDatabaseCoordinator

# Export
from relstorage._compat import MAX_TID # pylint:disable=unused-import

# pylint: disable=inherit-non-class,no-method-argument,no-self-argument
# pylint:disable=unexpected-special-method-signature
# pylint:disable=signature-differs

class IStorageCache(IDetachableMVCCDatabaseViewer):
    """
    A cache, as used by :class:`relstorage.interfaces.IRelStorage`.

    Implementations do not have to be thread-safe. Use :meth:`new_instance` to
    get an object to use concurrently.

    This cache is for current objects, as viewed at the ``highest_visible_tid``
    of this object or before. While it is possible to request older or newer
    revisions using ``loadSerial``, such accesses are less likely to produce cache
    hits.
    """

    def new_instance(before=None):
        """
        Create a new object to be used concurrently.

        If *before* is given, it is an integer TID, giving the *maximum*
        transaction that will be visible to this object (this only works
        correctly for history-preserving storages).

        .. caution::

           Failing to provide *before* for a historical connection
           will lead to consistency errors.
        """

    # TODO: Fill me in.

class IStorageCacheMVCCDatabaseCoordinator(IMVCCDatabaseCoordinator):
    """
    Specialized cache coordinator.
    """

    # TODO: Fill me in.

    def poll(cache, conn, cursor):
        """
        Poll for invalidations since the last taken state of the viewer.

        Update the state of the viewer and update the global state maintained
        in this coordinator.

        This should be the first thing done after the *conn* is opened or
        a transaction has begun: in fact, to ensure that the global
        state never accidentally goes backwards, no snapshot of the database
        should have been taken with the connection yet. The queries we execute
        in this method should establish the snapshot for the first time.
        """

    def stats():
        """
        Return a dictionary with interesting keys and values
        about the global state of this object.
        """

class IStateCache(Interface):
    """
    The methods we use to store state information.

    This interface is defined in terms of OID and TID *integers*;
    implementations (such as memcache) that only support string
    keys will need to convert.

    All return values for states return ``(state_bytes, tid_int)``.

    We use special methods where possible because those are slightly
    faster to invoke.
    """

    def __getitem__(oid_tid):
        """
        Given an (oid, tid) pair, return the cache data (state_bytes,
        tid_int) for that object.

        The returned *tid_int* must match the requested tid.

        If the ``(oid, tid)`` pair isn't in the cache, return None.

        A special tid value of None means that the client doesn't know the TID to ask
        for and wants the best available; they are then required to verify
        that the returned object is visible (that is, within the ranges of TIDs
        the MVCC state permits).
        """

    def get(oid_tid, peek=False): # pylint:disable=arguments-differ
        """
        As for :meth:`__getitem__`, but allows the caller to specify that this is
        a "peek" operation and shouldn't affect any internal bookkeeping
        such as statistics or most/least recently used lists.
        """

    def __setitem__(oid_tid, state_bytes_tid):
        """
        Store the *state_bytes_tid* (``(state_bytes, tid_int)``) for
        the ``(oid, tid)`` pair.

        Note that it does not necessarily mean that the key tid
        matches the value tid.

        Also note that if the object has been deleted, ``state_bytes``
        may be `None`.
        """

    def __delitem__(oid_tid):
        """
        Remove the data cached for the oid/tid pair.

        If no data is cached, this should do nothing.
        """

    def set_all_for_tid(tid_int, state_oid_iter):
        """
        Store the states for the ``(state, oid_int)`` pairs under
        ``(oid_int, tid_int)`` keys.

        The *state_oid_iter* is an **iteable** of ``(state, oid_int,
        prev_tid_int)`` pairs; this method may choose to buffer a
        limited amount of those pairs before passing them on to the
        underlying storage in a bulk group.

        As an **iterable**, *state_oid_iter* may be consumed multiple
        times.
        """

    def close():
        """
        Release external resources held by this object.

        This object may not be usable after this.
        """

    def new_instance():
        """
        Create an object sharing the same underlying data, but
        capable of operating independently, such as in a new thread.

        This may return the same object.
        """

    def release():
        """
        Like close, but intended to be called on child objects
        created for MVCC using a ``new_instance`` method.
        """

    def flush_all():
        """
        Clear cached data.
        """

    def invalidate_all(oids):
        """
        Remove all cached data for each of the oid integers in *oids*.
        """

class IPersistentCache(Interface):
    """
    A cache that can be persisted to a location on disk
    and later re-populated from that same location.
    """

    size = Attribute("The byte-size of the entries in the cache.")
    limit = Attribute("The upper bound of the byte-size that this cache should hold.")

    def save():
        """
        Save the cache to disk.
        """

    def restore():
        """
        Restore the cache from disk.
        """

    def zap_all():
        """
        Remove the cache from disk.
        """

class ILRUEntry(Interface):
    """
    An entry in an `ILRUCache`.

    This is a read-only object. The containing cache
    is in charge of all the attributes, and they must not
    be changed behind its back.

    """

    key = Attribute("The key for the entry")
    value = Attribute("The value for the entry")
    frequency = Attribute("The frequency of accesses to the entry.")
    weight = Attribute("The weight of the entry.")

class ILRUCache(Interface):
    """
    A container of cached keys and values and associated metadata,
    limited to containing a total weight less than some limit.

    The interface is mapping-like, and specified in terms of keys and
    values. For access to the additional stored metadata, different
    methods are defined.

    Values may be evicted when new ones are added.

    The cache may not store None values.

    The cache may have specific restrictions on the type and format
    of keys and values it accepts.
    """

    limit = Attribute("The maximim weight allowed.")
    weight = Attribute("The weight of the entries in the cache.")

    # Mapping-like methods.

    def __getitem__(key):
        """
        Get a value by key. If there is no value
        for the key, or it has been evicted, return None.

        This should be considered a hit on the key.

        This never results in raising a `KeyError`.
        """

    def __len__():
        """
        Count how many entries are in the cache.
        """
    def __contains__(key):
        "Is the key in the cache?"

    def __iter__():
        "Iterate the keys"

    def __setitem__(key, value):
        """
        Either set or update an entry.

        If the key already existed in the cache, then update its
        ``value`` to the new *value* and mark it as the most recently
        used.

        Otherwise, create a new entry for the key, setting it to the
        most recently used.

        This may evict other items.
        """

    def __delitem__(key):
        """
        Remove the entry from the cache.

        If it does not exist, it is an error.
        """

    ###
    # Cache-specific operations.
    ###

    def peek(key):
        """
        Similar to ``__getitem__``, but *does not* count
        as a hit on the key, merely returns a value if its present.
        """

    def values():
        """
        Iterate all the `ILRUEntry` values.
        """

    def add_MRUs(ordered_keys_and_values, return_count_only=False):
        """
        Add as many of the key/value pairs in *ordered_keys_and_values* as possible,
        without evicting any existing items.

        Returns the entries that were added, unless *return_count_only* is given,
        in which case it returns the count added instead. (This can save memory
        if the entries are not actually needed.)
        """

    def age_frequencies():
        """Call to periodically adjust the frequencies of items."""

class IGeneration(Interface):
    """
    A generation in a cache.
    """
    limit = Attribute("The maximim weight allowed.")
    __name__ = Attribute("The name of the generation")
    generation_number = Attribute("The number of the generation.")

    def __iter__():
        """
        Iterate the ILRUEntry objects in this generation,
        from most recent to least recently used.
        """

class IGenerationalLRUCache(ILRUCache):
    """
    The cache moves items between three generations, as determined by
    an admittance policy.

    * Items begin in *eden*, where they stay until eden grows too
      large.

    * When eden grows too large, the least recently used item is then
      (conceptually) moved to the *probation* ring. If this would make
      the probation ring too large, the *frequency* of the least
      recently used item from the probation ring is compared to the
      frequency of the incoming item. Only if the incoming item is more
      popular than the item it would force off the probation ring is it
      kept (and the probation item removed). Otherwise the eden item is
      removed.

    * When an item in probation is accessed, it is moved to the
      *protected* ring. The protected ring is the largest ring. When
      adding an item to it would make it too large, the least recently
      used item is demoted to probation, following the same rules as for
      eden.

    This cache only approximately follows its size limit. It may
    temporarily become larger.
    """

    eden = Attribute("The youngest generation.")
    protected = Attribute("The protected generation.")
    probation = Attribute("The probation generation.")
    generations = Attribute("Ordered list of generations, with 0 being NoSuchGeneration.")


class CacheCorruptedError(StorageError):
    """
    Raised when we detect internal cache corruption,
    outside of any particular transaction.
    """

class CacheConsistencyError(StorageError, TransientError):
    """
    Raised when we detect internal cache corruption, having to do with
    transactional consistency.

    This is probably a transient error. By the time it is raised, the
    faulty cache will have been cleared, and we can try again.
    """

class NoSuchGeneration(object):
    # For more specific error messages; if we get an AttributeError
    # on this object, it means we have corrupted the cache. We only
    # expect to wind up with this in generation 0.
    __name__ = 'NoSuchGeneration'
    def __init__(self, generation_number):
        self.__gen_num = generation_number

    def __getattr__(self, name):
        msg = "Generation %s has no attribute %r" % (self.__gen_num, name)
        raise AttributeError(msg)

class GenerationalCacheBase(object):
    # Percentage of our byte limit that should be dedicated
    # to the main "protected" generation
    _gen_protected_pct = 0.8
    # Percentage of our byte limit that should be dedicated
    # to the initial "eden" generation
    _gen_eden_pct = 0.1
    # Percentage of our byte limit that should be dedicated
    # to the "probationary"generation
    _gen_probation_pct = 0.1
    # By default these numbers add up to 1.0, but it would be possible to
    # overcommit by making them sum to more than 1.0. (For very small
    # limits, the rounding will also make them overcommit).

    #: A "mapping" between the __parent__ of an entry and the generation
    #: ring that holds it. (Indexing by ints is faster than a dictionary lookup
    #: especially on PyPy.) Initialize to an object that will report a bad lookup
    #: for the right generation. (The generator expression keeps us from leaking the
    #: indexing variable on Py2.)
    generations = tuple((NoSuchGeneration(i) for i in range(4)))

    def __init__(self, limit, eden, protected, probation):
        self.limit = limit
        self.eden = eden
        self.protected = protected
        self.probation = probation
        # Preserve the NoSuchGeneration initializers
        generations = list(GenerationalCacheBase.generations)
        for gen in (self.protected, self.probation, self.eden):
            generations[gen.generation_number] = gen
        self.generations = tuple(generations)



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os.path

from ..connmanager import AbstractConnectionManager

logger = __import__('logging').getLogger(__name__)


class Sqlite3ConnectionManager(AbstractConnectionManager):
    """
    SQLite doesn't really have isolation levels in the traditional
    sense; as far as that goes, it always operates in SERIALIZABLE
    mode. Instead, the connection's ``isolation_level`` parameter
    determines how autocommit behaves and the interaction between it
    at the SQLite and Python levels::

        - If it is ``None``, then the Python Connection object has nothing
          to do with transactions. SQLite operates in its default autocommit
          mode, beginning and ending a (read or write, as needed)
          transaction around every statement execution.

        - If it is set to IMMEDIATE or DEFERRED, than the Python
          Connection object watches each statement sent
          to ``Cursor.execute`` or ``Cursor.executemany`` to see
          if it's a DML statement (INSERT/UPDATE/DELETE/REPLACE, and
          prior to Python 3.6, basically anything else except a
          SELECT). If it is, and the sqlite level is not already
          in a transaction, then the Connection begins a transaction
          of the specified type before executing the statement. The
          Connection COMMITs when commit() is called or when a DDL
          statement is executed (prior to Python 3.6).

          For SELECT statements, the connection remains in its current state,
          which would be autocommit (read  committed) unless a transaction
          has been opened.

    Now, those are all write statements that theoretically would begin
    a write transaction and take a database lock, so it would seem
    like there's no difference between IMMEDIATE and DEFERRED. But
    there is: recall that temporary tables are actually in a temporary
    *database*, so when you write to one of those, sqlite's internal
    transaction locks only apply to it. A DEFERRED transaction thus
    allows other connections to write to their own temporary tables,
    while an IMMEDIATE transaction blocks other from writing to their
    temporaries.

    We thus tell Python SQLite to operate in DEFERRED mode; for load
    connections, we must explicitly execute the BEGIN to get a
    consistent snapshot of the database, but for store, we can let the
    Python Connection detect when we execute a write operation and
    begin the transaction then, letting SQLite upgrade the locks to
    explicit mode when we attempt a write to the main database.

    Taking an exclusive lock on the main database can be accomplished
    with any UPDATE statement, even one that doesn't match any actual
    rows.
    """

    # super copies these into isolation_load/store
    # We use 'EXCLUSIVE' for the load connection isolation but we don't
    # ever actually expect to get to that. It just makes problems
    # more apparent should that happen. EXCLUSIVE and IMMEDIATE do the same
    # thing in WAL mode, but in other modes, EXCLUSIVE also prevents
    # read access to the database. DEFERRED is the default if you just use
    # BEGIN TRANSACTION.
    isolation_serializable = 'EXCLUSIVE'
    isolation_read_committed = 'DEFERRED'

    def __init__(self, driver, pragmas, path, options):
        """
        :param dict pragmas: A map from string pragma name to string
            pragma value. These will be executed at connection open
            time. Except for WAL, and a few other critical values, we
            allow changing just about all the settings, letting the
            user turn off just about all the safety features so that
            the user can tune to their liking. The user can also set the
            ``max_page_count`` to operate as a quota system.
        """
        self.path = path
        self.keep_history = options.keep_history
        self.pragmas = pragmas.copy()
        # Things that we really want to be on.
        nice_to_have_pragmas = {
            'foreign_keys': 1
        }
        nice_to_have_pragmas.update(pragmas)
        self.pragmas = nice_to_have_pragmas
        # XXX: Why do we override the default for this?
        self.pragmas['journal_size_limit'] = None
        super(Sqlite3ConnectionManager, self).__init__(options, driver)

        assert self.isolation_load == type(self).isolation_serializable
        assert self.isolation_store == type(self).isolation_read_committed


    def open(self,
             isolation=None,
             read_only=False,
             deferrable=False,
             replica_selector=None,
             application_name=None,
             **kwargs):

        if not os.path.exists(os.path.dirname(self.path)) and self.options.create_schema:
            os.makedirs(os.path.dirname(self.path))

        conn = self.driver.connect_to_file(
            self.path,
            query_only=read_only,
            timeout=self.options.commit_lock_timeout,
            quick_check=False,
            isolation_level=isolation,
            extra_pragmas=self.pragmas)
        cur = conn.cursor()
        return conn, cur

    def _do_open_for_load(self):
        conn, cur = self.open(
            isolation=self.isolation_load,
            read_only=True
        )
        # If we don't explicitly begin a transaction block, we're in
        # autocommit-mode for SELECT, which is essentially REPEATABLE READ, i.e.,
        # our MVCC state moves forward each time we read and we don't have a consistent
        # view.
        cur.execute('BEGIN DEFERRED TRANSACTION')
        return conn, cur

    def restart_load(self, conn, cursor, needs_rollback=True):
        needs_rollback = True
        super(Sqlite3ConnectionManager, self).restart_load(conn, cursor, needs_rollback)
        assert not conn.in_transaction
        # See _do_open_for_load.
        cursor.execute('BEGIN DEFERRED TRANSACTION')

    def open_for_pre_pack(self):
        # This operates in auto-commit mode.
        conn, cur = self.open(isolation=None)
        assert conn.isolation_level is None
        return conn, cur

    def open_for_pack_lock(self):
        # We don't use a pack lock.
        return (None, None)



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from zope.interface import alsoProvides

from relstorage.tests import MockDriver
from ..adapter import OracleAdapter as BaseAdapter

from ... import interfaces
from ...tests import test_adapter

class Adapter(BaseAdapter):
    # We don't usually have cx_oracle installed, so
    # fake it.

    def _select_driver(self, options=None):
        d = MockDriver()
        d.connect = lambda *args, **kwargs: None
        d.cursor = lambda conn, server_side=False: None
        d.NUMBER = None
        d.BLOB = None
        d.BINARY = None
        d.STRING = None
        d.Binary = None
        d.use_replica_exceptions = ()
        d.binary_column_as_state_type = lambda b: b
        d.binary_column_as_bytes = lambda b: b
        d.__name__ = 'cx_Oracle'
        alsoProvides(d, interfaces.IDBDriver)
        return d

class TestAdapter(test_adapter.AdapterTestBase):

    def _makeOne(self, options):
        return Adapter(options)



# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from relstorage.tests import TestCase
from relstorage.adapters.tests.test_drivers import IDBDriverSupportsCriticalTestMixin
from ..import mysqldb



class GeventMySQLdbDriver(mysqldb.GeventMySQLdbDriver):
    GeventConnection = None

    @classmethod
    def _get_connection_class(cls):
        if cls.GeventConnection is None:
            from .. import _mysqldb_gevent
            class GeventConnection(_mysqldb_gevent.AbstractConnection):
                class GeventCursor(object):
                    def __init__(self, conn):
                        self.connection = conn
                    execute = lambda self, _q, _a: None
                    fetchall = lambda self: []
                    nextset = lambda self: None

                def __init__(self, *args, **kwargs): # pylint:disable=super-init-not-called
                    self.connect_args = args
                    self.conect_kwargs = kwargs

                _direct_commit = _direct_rollback = lambda self: None

                def cursor(self):
                    return self.GeventCursor(self)
            cls.GeventConnection = GeventConnection
        return cls.GeventConnection

class TestGeventMySQLdb(IDBDriverSupportsCriticalTestMixin,
                        TestCase):

    def _makeOne(self):
        try:
            return GeventMySQLdbDriver()
        except ImportError as e:
            self.skipTest(e)

    def test_callproc_exit_critical_phase(self):
        driver = self._makeOne()
        conn = driver.connect()

        driver.enter_critical_phase_until_transaction_end(conn, None)

        driver.callproc_multi_result(
            conn.cursor(),
            'do_something()',
            (),
            exit_critical_phase=True
        )

        self.assertFalse(driver.is_in_critical_phase(conn, None))



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from ..adapter import MySQLAdapter as Adapter
from ...tests import test_adapter

class MockVersionDetector(object):

    def supports_good_stored_procs(self, _cursor):
        return True

class MockConnection(object):

    cursor = None

class MockDriver(object):

    last_callproc_args = None
    last_callproc_kwargs = None
    last_callproc_multi_result = ((),)

    def callproc_multi_result(self, _cur, *args, **kwargs):
        self.last_callproc_args = args
        self.last_callproc_kwargs = kwargs
        return self.last_callproc_multi_result

class TestAdapter(test_adapter.AdapterTestBase):

    def _makeOne(self, options):
        return Adapter(options=options)

    def _test_lock_database_and_move_ends_critical_section_on_commit(self, commit):
        adapter = self._makeOne(None)
        driver = adapter.driver = MockDriver()
        driver.last_callproc_multi_result = (([42],),)
        adapter.version_detector = MockVersionDetector()

        result = adapter.lock_database_and_move(
            MockConnection(), None,
            None,
            (),
            commit=commit
        )

        self.assertEqual(result, (42, "-"))

        self.assertEqual(
            driver.last_callproc_kwargs,
            {'exit_critical_phase': commit}
        )

    def test_lock_database_and_move_ends_critical_section_on_commit(self):
        self._test_lock_database_and_move_ends_critical_section_on_commit(True)

    def test_lock_database_and_move_keeps_critical_section_wo_commit(self):
        self._test_lock_database_and_move_ends_critical_section_on_commit(False)



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from __future__ import absolute_import

import logging
from relstorage._compat import iteritems

from ..scriptrunner import ScriptRunner

log = logging.getLogger(__name__)

_stmt_cache = {}

def _format_to_named(stmt):
    """
    Convert '%s' pyformat strings to :n numbered
    strings. Intended only for static strings.

    This is legacy. Replace strings that use this with SQL statements
    constructed from the schema.
    """
    import re
    from relstorage._compat import intern

    try:
        return _stmt_cache[stmt]
    except KeyError:
        matches = []

        def replace(_match):
            matches.append(None)
            return ':%d' % len(matches)
        new_stmt = intern(re.sub('%s', replace, stmt))
        _stmt_cache[stmt] = new_stmt

        return new_stmt

class OracleScriptRunner(ScriptRunner):

    script_vars = dict(ScriptRunner.script_vars)
    script_vars.update({
        'TRUE':         "'Y'",
        'FALSE':        "'N'",
        'oid':          ':oid',
        'tid':          ':tid',
        'pack_tid':     ':pack_tid',
        'undo_tid':     ':undo_tid',
        'self_tid':     ':self_tid',
        'min_tid':      ':min_tid',
        'max_tid':      ':max_tid',
        # Oracle won't accept ORDER BY clauses inside
        # the subquery of an IN
        'INNER_ORDER_BY': ''
    })

    def run_script_stmt(self, cursor, generic_stmt, generic_params=()):
        """Execute a statement from a script with the given parameters.

        params should be either an empty tuple (no parameters) or
        a map.
        """
        generic_stmt = generic_stmt.format(**self.format_vars)
        # We can't quote "transaction", but we have to for sqlite.
        generic_stmt = generic_stmt.replace(' "transaction"', ' transaction')

        if generic_params and isinstance(generic_params, tuple):
            generic_stmt = _format_to_named(generic_stmt) # Unnamed params become numbered.
        if generic_params and isinstance(generic_params, dict):
            # Oracle raises ORA-01036 if the parameter map contains extra keys,
            # so filter out any unused parameters.
            tracker = TrackingMap(self.script_vars)
            stmt = generic_stmt % tracker
            used = tracker.used
            params = {}
            for k, v in iteritems(generic_params):
                if k in used:
                    params[k] = v

        else:
            stmt = generic_stmt % self.script_vars
            params = ()

        if generic_params and isinstance(generic_params, tuple):
            params = generic_params
        __traceback_info__ = stmt

        try:
            cursor.execute(stmt, params)
        except:
            log.warning("script statement failed: %r; parameters: %r",
                        stmt, params)
            raise

    def run_many(self, cursor, stmt, items):
        """Execute a statement repeatedly.  Items should be a list of tuples.

        stmt should use '%s' parameter format.
        """
        cursor.executemany(_format_to_named(stmt), items)


class TrackingMap(object):
    """Provides values for keys while tracking which keys are accessed."""

    def __init__(self, source):
        self.source = source
        self.used = set()

    def __getitem__(self, key):
        self.used.add(key)
        return self.source[key]


class CXOracleScriptRunner(OracleScriptRunner):

    def __init__(self, driver):
        self.driver = driver

    def new_instance(self):
        return type(self)(self.driver)


    def _read_lob(self, value):
        """Handle an Oracle LOB by returning its byte stream.

        Returns other objects unchanged.
        """
        if isinstance(value, self.driver.LOB):
            return value.read()
        return value

    def run_lob_stmt(self, cursor, stmt, args=(), default=None):
        """
        Execute a statement and return one row with all LOBs
        inline.

        Returns the value of the default parameter if the result was
        empty.

        The statement can either be a string, or a CompiledQuery
        object.
        """

        try:
            if hasattr(stmt, 'execute'):
                stmt.execute(cursor, args)
            else:
                cursor.execute(stmt, args)
            rows = cursor.fetchall()
        except self.driver.DatabaseError as e:
            # ORA-01406: fetched column value was truncated
            error = e.args[0]

            if ((isinstance(error, str) and not error.endswith(' 1406'))
                    or error.code != 1406):
                raise
            # Execute the query, but alter it slightly without
            # changing its meaning, so that the query cache
            # will see it as a statement that has to be compiled
            # with different output type parameters.
            oth = cursor.connection.outputtypehandler
            cursor.connection.outputtypehandler = None
            try:
                cursor.execute(stmt + ' ', args)
                rows = [
                    tuple([self._read_lob(x) for x in row])
                    for row in cursor
                ]
            finally:
                cursor.connection.outputtypehandler = oth

        assert len(rows) in (0, 1)
        return rows[0] if rows else default



# -*- coding: utf-8 -*-
"""
Script to run zope.testrunner in a gevent monkey-patched environment.

Using ``python -m gevent.monkey zope-testrunner ...`` is insufficient.

This is because up through 1.5a2 there is a serious bug in the way the
monkey-patcher patches the spawned process. The net effect is that the gevent
threadpool isn't functional.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import gevent.monkey
gevent.monkey.patch_all()
# pylint:disable=wrong-import-position, wrong-import-order
import sys

from zope.testrunner import run

sys.argv[:] = [
    'zope-testrunner',
    '--test-path=src',
    '-vvv',
    '--color',
] + sys.argv[1:]
print(sys.argv)
run()



##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Implements restoring from one database to another.

"""
from __future__ import absolute_import
from __future__ import print_function

from ZODB.POSException import StorageTransactionError

from ZODB.utils import u64 as bytes8_to_int64

from .vote import DatabaseLockedForTid
from .vote import HistoryFree as HFVoteFactory
from .vote import HistoryPreserving as HPVoteFactory


class Restore(object):
    """
    A type of begin state that wraps another begin state and adds the methods needed to
    restore or commit to a particular tid.
    """

    # batcher: An object that accumulates store operations
    # so they can be executed in batch (to minimize latency).
    batcher = None

    # _batcher_row_limit: The number of rows to queue before
    # calling the database.
    batcher_row_limit = 100

    # Methods from the underlying begin implementation we need to
    # expose.
    _COPY_ATTRS = (
        'store',
        'checkCurrentSerialInTransaction',
        'deletObject',
        'undo',
        'tpc_vote',
        'tpc_abort',
        'no_longer_stale',
        'shared_state',
    )

    def __init__(self, begin_state, committing_tid, status):
        # type: (relstorage.storage.tpc.begin.AbstractBegin, DatabaseLockedForTid, str) -> None

        # This is an extension we use for copyTransactionsFrom;
        # it is not part of the IStorage API.
        assert committing_tid is not None
        self.wrapping = begin_state # type: AbstractTPCState

        # hold the commit lock and add the transaction now.
        # This will prevent anyone else from modifying rows
        # other than this transaction. We currently avoid the temp tables,
        # though, so if we do multiple things in a restore transaction,
        # we could still wind up with locking issues (I think?)
        adapter = begin_state.shared_state.adapter
        cursor = begin_state.shared_state.store_connection.cursor
        packed = (status == 'p')
        try:
            committing_tid_lock = DatabaseLockedForTid.lock_database_for_given_tid(
                committing_tid, packed,
                cursor, adapter, begin_state.ude
            )
        except:
            begin_state.shared_state.store_connection.drop()
            raise

        # This is now only used for restore()
        self.batcher = batcher = adapter.mover.make_batcher(
            cursor,
            self.batcher_row_limit)

        for name in self._COPY_ATTRS:
            try:
                meth = getattr(begin_state, name)
            except AttributeError:
                continue
            else:
                setattr(self, name, meth)

        # Arrange for voting to store our batch too, since
        # the mover is unaware of it.
        factory = begin_state.tpc_vote_factory
        assert factory is HFVoteFactory or factory is HPVoteFactory
        def tpc_vote_factory(begin_state,
                             _f=_HFVoteFactory if factory is HFVoteFactory else _HPVoteFactory,
                             _c=committing_tid_lock,
                             _b=batcher):
            return _f(begin_state, _c, _b)
        begin_state.tpc_vote_factory = tpc_vote_factory
        begin_state.shared_state.temp_storage = _TempStorageWrapper(
            begin_state.shared_state.temp_storage)

    def restore(self, oid, this_tid, data, prev_txn, transaction):
        # Similar to store() (see comments in FileStorage.restore for
        # some differences), but used for importing transactions.
        # Note that *data* can be None.
        # The *prev_txn* "backpointer" optimization/hint is ignored.
        #
        # pylint:disable=unused-argument
        state = self.wrapping
        if transaction is not state.transaction:
            raise StorageTransactionError(self, transaction)

        adapter = state.shared_state.adapter
        cursor = state.shared_state.store_connection.cursor
        assert cursor is not None
        oid_int = bytes8_to_int64(oid)
        tid_int = bytes8_to_int64(this_tid)

        # Save the `data`.  Note that `data` can be None.
        # Note also that this doesn't go through the cache.
        state.shared_state.temp_storage.max_restored_oid = max(
            state.shared_state.temp_storage.max_restored_oid,
            oid_int)
        # TODO: Make it go through the cache, or at least the same
        # sort of queing thing, so that we can do a bulk COPY.
        # The way we do it now complicates restoreBlob() and it complicates voting.
        adapter.mover.restore(
            cursor, self.batcher, oid_int, tid_int, data)

    def restoreBlob(self, oid, serial, data, blobfilename, prev_txn, txn):
        self.restore(oid, serial, data, prev_txn, txn)
        state = self.wrapping
        # Restoring the entry for the blob MAY have used the batcher, and
        # we're going to want to foreign-key off of that data when
        # we add blob chunks (since we skip the temp tables).
        # Ideally, we wouldn't need to flush the batcher here
        # (we'd prefer having DEFERRABLE INITIALLY DEFERRED FK
        # constraints, but as-of 8.0 MySQL doesn't support that.)
        self.batcher.flush()
        cursor = state.shared_state.store_connection.cursor
        state.shared_state.blobhelper.restoreBlob(cursor, oid, serial, blobfilename)

class _TempStorageWrapper(object):

    __slots__ = (
        'temp_storage',
        'max_restored_oid',
    )

    def __init__(self, temp_storage):
        self.temp_storage = temp_storage
        self.max_restored_oid = 0

    @property
    def max_stored_oid(self):
        return max(self.temp_storage.max_stored_oid, self.max_restored_oid)

    def __len__(self):
        return len(self.temp_storage)

    def __iter__(self):
        return iter(self.temp_storage)

    def __getattr__(self, name):
        return getattr(self.temp_storage, name)

class _VoteFactoryMixin(object):
    __slots__ = ()

    def __init__(self, begin_state, committing_tid_lock, batcher):
        # type: (Restore, Optional[DatabaseLockedForTid], Any) -> None
        super(_VoteFactoryMixin, self).__init__(begin_state)
        # pylint:disable=assigning-non-slot
        self.committing_tid_lock = committing_tid_lock
        self.batcher = batcher

    def _flush_temps_to_db(self, cursor):
        super(_VoteFactoryMixin, self)._flush_temps_to_db(cursor)
        self.batcher.flush()


class _HFVoteFactory(_VoteFactoryMixin, HFVoteFactory):
    __slots__ = ('batcher',)


class _HPVoteFactory(_VoteFactoryMixin, HPVoteFactory):
    __slots__ = ('batcher',)



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Stats implementations
"""

import abc

from zope.interface import implementer

from .._compat import ABC
from .interfaces import IDBStats
from ._util import DatabaseHelpersMixin
from .schema import Schema
from .sql import func


@implementer(IDBStats)
class AbstractStats(DatabaseHelpersMixin, ABC):

    def __init__(self, connmanager, keep_history):
        self.connmanager = connmanager
        self.keep_history = keep_history

    _get_object_count_query = Schema.all_current_object.select(
        func.count()
    )


    def get_object_count(self):
        """Returns the approximate number of objects in the database"""
        conn, cursor = self.connmanager.open_for_load()
        try:
            self._get_object_count_query.execute(cursor)
            return cursor.fetchone()[0]
        finally:
            self.connmanager.close(conn, cursor)

    @abc.abstractmethod
    def get_db_size(self):
        """Returns the approximate size of the database in bytes"""
        raise NotImplementedError()


    def large_database_change(self):
        """
        Call this when the database has changed substantially,
        and it would be a good time to perform any updates or
        optimizations.
        """



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Interfaces for RelStorage implementation components.

These interfaces aren't meant to be considered public, they exist to
serve as documentation and for validation of RelStorage internals.

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from zope.interface import Interface
from zope.interface import Attribute

from transaction.interfaces import TransientError
from ZODB.POSException import StorageTransactionError
from ZODB.POSException import ReadConflictError

# pylint:disable=inherit-non-class,no-self-argument,no-method-argument
# pylint:disable=too-many-ancestors

class StorageDisconnectedDuringCommit(StorageTransactionError, TransientError):
    """
    We lost a connection to the database during a commit. Probably restarting
    the transaction will re-establish the connection.
    """

class IStaleAware(Interface):
    """
    An object that can transition between a state of normalcy and being stale.

    In the stale state, calling the object raises an exception.
    """

    def stale(stale_error):
        """
        Produce a new object that is the stale version of this object
        and return it.

        :return: Another `IStaleAware`.
        """

    def no_longer_stale():
        """
        Produce a new (or the original) object that is in the normal state
        and return it.

        :return: Another `IStaleAware`.
        """

class VoteReadConflictError(ReadConflictError):
    """
    A read conflict (from Connection.readCurrent()) that wasn't
    detected until the storage voted.
    """

class ITPCState(Interface):
    """
    An object representing the current state (phase) of the two-phase commit protocol,
    and how to transition between it and other phases.

    The initial state is always :class:`ITPCStateNotInTransaction`
    """

    transaction = Attribute("The *transaction* object from ZODB.")

    initial_state = Attribute("The ITPCStateNotInTransaction that started the process.")

    def tpc_abort(transaction, force=False):
        """
        Clear any used resources, and return the object
        representing :class:`ITPCStateNotInTransaction`.

        :param transaction: The transaction object from ZODB.
           If this is not the current transaction, does nothing
           unless *force* is true.
        :keyword bool force: Whether to forcibly abort the transaction.
           If this is true, then it doesn't matter if the *transaction* parameter
           matches or not. Also, underlying RDBMS connections should also be closed
           and discarded.
        :return: The previous `ITPCStateNotInTransaction`.
        """


class ITPCStateNotInTransaction(ITPCState):
    """
    The "idle" state.

    In this state, no store connection is available,
    and the *transaction* is always `None`.

    Because ZODB tests this, this method has to define
    a bunch of methods that are also defined by various other states.
    These methods should raise ``StorageTransactionError``, or
    ``ReadOnlyError``, as appropriate.

    Implementations of this interface should be false in a boolean context
    to easily permit testing whether a TPC phase is active or not.
    """

    last_committed_tid_int = Attribute(
        """
        The TID of the last transaction committed to get us to this state.

        Initially, this is 0. In the value returned from :meth:`ITPCPhaseVoting.tpc_finish`,
        it is the TID just committed.
        """
    )

    def tpc_begin(storage, transaction):
        """
        Enter the two-phase commit protocol.

        :return: An implementation of :class:`ITPCStateBegan`.
        """

    def tpc_finish(*args, **kwargs):
        """
        Raises ``StorageTranasctionError``
        """

    def tpc_vote(*args, **kwargs):
        """
        As for `tpc_finish`.
        """

    def store(*args, **kwargs):
        """
        Raises ``ReadOnlyError`` or ``StorageTranasctionError``
        """

    restore = deleteObject = undo = restoreBlob = store


class ITPCStateDatabaseAvailable(ITPCState, IStaleAware):
    """
    A state where the writable database connection is available.
    """

    store_connection = Attribute("The IManagedStoreConnection in use.")


class ITPCStateBegan(ITPCStateDatabaseAvailable):
    """
    The first phase where the database is available for storage.

    Storing objects, restoring objects, storing blobs, deleting objects,
    all happen in this phase.
    """

    def tpc_vote(storage, transaction):
        """
        Enter the voting state.

        :return: An implementation of :class:`ITPCStateVoting`
        """

class ITPCStateBeganHP(ITPCStateBegan):
    """
    The extra methods that are available for history-preserving
    storages.
    """

class ITPCStateVoting(ITPCStateDatabaseAvailable):
    """
    The phase where voting happens. This follows the beginning phase.
    """

    invalidated_oids = Attribute(
        """An iterable of OID bytes, returned from the storage's ``tpc_vote`` method.

        The Connection will ghost all cached objects in this iterable. This includes
        things things during conflict resolution or undo.
        """
    )

    def tpc_finish(storage, transaction, f=None):
        """
        Finish the transaction.

        :return: The next implementation of :class:`ITPCPhaseNotInTransaction`
        """



##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""ZConfig directive implementations for binding RelStorage to Zope"""
from __future__ import absolute_import

from relstorage.options import Options
from relstorage.storage import RelStorage

logger = __import__('logging').getLogger(__name__)

class BaseConfig(object):

    def __init__(self, config):
        self.config = config
        self.name = config.getSectionName()

class RelStorageFactory(BaseConfig):
    """Open a storage configured via ZConfig"""
    def open(self):
        config = self.config
        # Hoist the driver setting to the section we really want it.
        config.driver = config.adapter.config.driver
        # But don't remove it or otherwise mutate the config object;
        # that would prevent us from being correctly opened again.
        #config.adapter.config.driver = None
        options = Options.copy_valid_options(config)
        options.adapter = config.adapter
        # The adapter factories may modify the global options (or raise an exception)
        # if something at the top-level is specifically not allowed based on
        # their configuration.
        adapter = config.adapter.create(options)
        return RelStorage(adapter, name=config.name, options=options)


class PostgreSQLAdapterFactory(BaseConfig):
    def create(self, options):
        from .adapters.postgresql import PostgreSQLAdapter
        return PostgreSQLAdapter(
            dsn=self.config.dsn,
            options=options)


class OracleAdapterFactory(BaseConfig):
    def create(self, options):
        from .adapters.oracle import OracleAdapter
        config = self.config
        return OracleAdapter(
            user=config.user,
            password=config.password,
            dsn=config.dsn,
            options=options)


class MySQLAdapterFactory(BaseConfig):
    def create(self, options):
        from .adapters.mysql import MySQLAdapter
        params = {}
        for key in self.config.getSectionAttributes():
            if key == 'driver':
                continue
            value = getattr(self.config, key)
            if value is not None:
                params[key] = value
        return MySQLAdapter(options=options, **params)


class SQLitePragmas(BaseConfig):

    @property
    def pragmas(self):
        pragmas = self.config.pragmas
        # The keys could be repeated so the values are lists
        pragmas = {k: v for k, (v,) in pragmas.items()}
        # Ignore busy_timeout: That's the same thing as commit-lock-timeout
        pragmas.pop('busy_timeout', None)
        for attr in self.config.getSectionAttributes():
            if attr == 'pragmas':
                continue
            val = getattr(self.config, attr)
            if val is not None:
                pragmas[attr] = val
        return pragmas

class Sqlite3AdapterFactory(BaseConfig):
    def create(self, options):
        from .adapters.sqlite.adapter import Sqlite3Adapter
        if options.cache_local_dir:
            # A persistent cache makes absolutely no sense.
            # Disable.
            logger.info("Ignoring cache-local-dir setting.")
            del options.cache_local_dir
        if self.config.pragmas:
            pragmas = self.config.pragmas.pragmas
        else:
            pragmas = {}
        return Sqlite3Adapter(
            self.config.data_dir,
            pragmas=pragmas,
            options=options)



# -*- coding: utf-8 -*-
"""
Helpers for cache testing.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from zope import interface

from relstorage.cache.cache import PyCache as _BaseCache
from relstorage.cache.local_client import LocalClient as _BaseLocalClient

from relstorage.tests import MockOptions
from relstorage.tests import MockCursor
from relstorage.tests import MockConnectionManager
from relstorage.tests import MockAdapter

class MockOptionsWithFakeMemcache(MockOptions):
    cache_module_name = 'relstorage.tests.fakecache'
    cache_servers = 'host:9999'


class Cache(object):
    # Tweak the generation sizes to match what we developed the tests with
    _gen_protected_pct = 0.8
    _gen_eden_pct = 0.1
    _gen_probation_pct = 0.1
    _dict_type = dict


    def __init__(self, byte_limit):
        self.__cache = _BaseCache(
            byte_limit * self._gen_eden_pct,
            byte_limit * self._gen_protected_pct,
            byte_limit * self._gen_probation_pct
        )
        interface.alsoProvides(self, interface.providedBy(self.__cache))

    def __getitem__(self, key):
        return self.__cache[key]

    def __len__(self):
        return len(self.__cache)

    def __setitem__(self, k, v):
        self.__cache[k] = v

    def __delitem__(self, k):
        del self.__cache[k]

    def __iter__(self):
        return iter(self.__cache)

    def __contains__(self, k):
        return k in self.__cache

    def __getattr__(self, name):
        return getattr(self.__cache, name)

    @property
    def size(self):
        return self.__cache.weight

class LocalClient(_BaseLocalClient):
    _cache_type = Cache

    def __init__(self, options, prefix=None):
        if isinstance(options, int):
            options = MockOptions(cache_local_mb=options)
        super(LocalClient, self).__init__(options, prefix=prefix)

def list_lrukeys(mapping, generation_name):
    # Remember, these lists will be from LRU to MRU
    return [e.key for e in getattr(mapping, generation_name)]

def list_lrufreq(mapping, generation_name):
    return [e.frequency for e in getattr(mapping, generation_name)]



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""More recovery and iterator tests."""

# This is copied from ZODB.tests.RecoveryStorage and expanded to fit
# history-free storages.
# pylint:disable=no-member,too-many-locals

import itertools
import time

import transaction
import ZODB.POSException
from ZODB import DB
from ZODB.Connection import TransactionMetaData as Transaction
from ZODB.blob import is_blob_record
from ZODB.serialize import referencesf
from ZODB.tests.StorageTestBase import MinPO
from ZODB.tests.StorageTestBase import zodb_pickle

from relstorage.interfaces import IRelStorage
from relstorage._compat import PY3
from relstorage._compat import iteritems

if not PY3:
    zip = itertools.izip

class IteratorDeepCompare(object):

    def compare(self, src, dest):
        # override this for storages that truncate on restore (because
        # they do not store history).
        self.compare_exact(src, dest)

    def compare_exact(self, storage1, storage2):
        """Confirm that storage1 and storage2 contain equivalent data"""
        eq = self.assertEqual
        missing = object()
        # len(storage) is only required to be approximate and for "informational purposes."
        # So we cannot use it as a hard check.
        # self.assertEqual(len(storage1), len(storage2))
        iter1 = self._closing(storage1.iterator())
        iter2 = self._closing(storage2.iterator())
        for txn1, txn2 in zip(iter1, iter2):
            eq(txn1.tid, txn2.tid)
            eq(txn1.status, txn2.status)
            eq(txn1.user, txn2.user)
            eq(txn1.description, txn2.description)

            # b/w compat on the 'extension' attribute
            e1 = getattr(txn1, 'extension', missing)
            if e1 is missing:
                # old attribute name
                e1 = txn1._extension
            e2 = getattr(txn2, 'extension', missing)
            if e2 is missing:
                # old attribute name
                e2 = txn2._extension
            eq(e1, e2)

            # compare the objects in the transaction, but disregard
            # the order of the objects and any duplicated records
            # since those are not important.
            recs1 = {r.oid: r for r in txn1}
            recs2 = {r.oid: r for r in txn1}
            eq(len(recs1), len(recs2))
            recs1 = sorted(iteritems(recs1))
            recs2 = sorted(iteritems(recs2))
            recs2.sort()
            for (_oid1, rec1), (_oid2, rec2) in zip(recs1, recs2):
                eq(rec1.oid, rec2.oid)
                eq(rec1.tid, rec2.tid)
                eq(rec1.data, rec2.data)
                if is_blob_record(rec1.data):
                    try:
                        fn1 = storage1.loadBlob(rec1.oid, rec1.tid)
                    except ZODB.POSException.POSKeyError:
                        self.assertRaises(
                            ZODB.POSException.POSKeyError,
                            storage2.loadBlob, rec1.oid, rec1.tid)
                    else:
                        fn2 = storage2.loadBlob(rec1.oid, rec1.tid)
                        self.assertNotEqual(fn1, fn2)
                        with open(fn1, 'rb') as f1, open(fn2, 'rb') as f2:
                            eq(f1.read(), f2.read())

        # Make sure there are no more records left in txn1 and txn2, meaning
        # they were the same length
        try:
            next(iter1)
        except (IndexError, StopIteration):
            pass
        else:
            self.fail("storage1 has more records")

        try:
            __traceback_info__ = next(iter2)
        except (IndexError, StopIteration):
            pass
        else:
            self.fail("storage2 has more records")

        iter1.close()
        iter2.close()

    def compare_truncated(self, src, dest):
        """Confirm that dest is a truncated copy of src.

        The copy process should have dropped all old revisions of objects
        in src.  Also note that the dest does not retain transaction
        metadata.
        """

        src_objects = {}  # {oid: (tid, data, blob or None)}
        for txn in src.iterator():
            for rec in txn:
                if is_blob_record(rec.data):
                    try:
                        fn = src.loadBlob(rec.oid, rec.tid)
                    except ZODB.POSException.POSKeyError:
                        blob = None
                    else:
                        blob = open(fn, 'rb').read()
                else:
                    blob = None
                src_objects[rec.oid] = (rec.tid, rec.data, blob)

        unchecked = set(src_objects)
        for txn in dest.iterator():
            for rec in txn:
                if is_blob_record(rec.data):
                    try:
                        fn = dest.loadBlob(rec.oid, rec.tid)
                    except ZODB.POSException.POSKeyError:
                        blob = None
                    else:
                        blob = open(fn, 'rb').read()
                else:
                    blob = None
                dst_object = (rec.tid, rec.data, blob)
                src_object = src_objects[rec.oid]
                self.assertEqual(src_object, dst_object)
                unchecked.remove(rec.oid)

        self.assertEqual(len(unchecked), 0)


class BasicRecoveryStorage(IteratorDeepCompare):

    # Requires a setUp() that creates a self._dst destination storage
    def checkSimpleRecovery(self):
        oid = self._storage.new_oid()
        revid = self._dostore(oid, data=11)
        revid = self._dostore(oid, revid=revid, data=12)
        revid = self._dostore(oid, revid=revid, data=13)
        self._dst.copyTransactionsFrom(self._storage)
        self.compare(self._storage, self._dst)

    def checkPackWithGCOnDestinationAfterRestore(self):
        raises = self.assertRaises
        closing = self._closing
        __traceback_info__ = self._storage, self._dst
        db = closing(DB(self._storage))
        conn = closing(db.open())
        root = conn.root()
        root.obj = obj1 = MinPO(1)
        txn = transaction.get()
        txn.note(u'root -> obj')
        txn.commit()
        root.obj.obj = obj2 = MinPO(2)
        txn = transaction.get()
        txn.note(u'root -> obj -> obj')
        txn.commit()
        del root.obj
        txn = transaction.get()
        txn.note(u'root -X->')
        txn.commit()

        storage_last_tid = conn._storage.lastTransaction()
        self.assertEqual(storage_last_tid, root._p_serial)

        # Now copy the transactions to the destination
        self._dst.copyTransactionsFrom(self._storage)
        self.assertEqual(self._dst.lastTransaction(), storage_last_tid)
        # If the source storage is a history-free storage, all
        # of the transactions are now marked as packed in the
        # destination storage.  To trigger a pack, we have to
        # add another transaction to the destination that is
        # not packed.
        db2 = closing(DB(self._dst))
        tx_manager = transaction.TransactionManager(explicit=True)
        conn2 = closing(db2.open(tx_manager))
        txn = tx_manager.begin()
        root2 = conn2.root()
        root2.extra = 0
        txn.note(u'root.extra = 0')
        txn.commit()

        dest_last_tid = conn2._storage.lastTransaction()
        self.assertGreater(dest_last_tid, storage_last_tid)
        self.assertEqual(dest_last_tid, root2._p_serial)

        # Now pack the destination.
        from ZODB.utils import u64 as bytes8_to_int64
        if IRelStorage.providedBy(self._dst):
            packtime = bytes8_to_int64(storage_last_tid)
        else:
            from persistent.timestamp import TimeStamp
            packtime = TimeStamp(dest_last_tid).timeTime() + 2
        self._dst.pack(packtime, referencesf)
        # And check to see that the root object exists, but not the other
        # objects.
        __traceback_info__ += (packtime,)
        _data, _serial = self._dst.load(root._p_oid, '')
        raises(KeyError, self._dst.load, obj1._p_oid, '')
        raises(KeyError, self._dst.load, obj2._p_oid, '')

    def checkRestoreAfterDoubleCommit(self):
        oid = self._storage.new_oid()
        revid = b'\0' * 8
        data1 = zodb_pickle(MinPO(11))
        data2 = zodb_pickle(MinPO(12))
        # Begin the transaction
        t = Transaction()
        try:
            self._storage.tpc_begin(t)
            # Store an object
            self._storage.store(oid, revid, data1, '', t)
            # Store it again
            self._storage.store(oid, revid, data2, '', t)
            # Finish the transaction
            self._storage.tpc_vote(t)
            self._storage.tpc_finish(t)
        except:
            self._storage.tpc_abort(t)
            raise
        self._dst.copyTransactionsFrom(self._storage)
        self.compare(self._storage, self._dst)


class UndoableRecoveryStorage(BasicRecoveryStorage):
    """These tests require the source storage to be undoable"""

    def checkRestoreAcrossPack(self):
        closing = self._closing
        db = closing(DB(self._storage))
        c = closing(db.open())
        r = c.root()
        r["obj1"] = MinPO(1)
        transaction.commit()
        r["obj2"] = MinPO(1)
        transaction.commit()

        self._dst.copyTransactionsFrom(self._storage)
        self._dst.pack(time.time(), referencesf)

        self._undo(self._storage.undoInfo()[0]['id'])

        # copy the final transaction manually.  even though there
        # was a pack, the restore() ought to succeed.
        it = closing(self._storage.iterator())
        # Get the last transaction and its record iterator. Record iterators
        # can't be accessed out-of-order, so we need to do this in a bit
        # complicated way:
        final = None
        for final in it:
            records = list(final)

        self._dst.tpc_begin(final, final.tid, final.status)
        for r in records:
            self._dst.restore(r.oid, r.tid, r.data, '', r.data_txn,
                              final)
        self._dst.tpc_vote(final)
        self._dst.tpc_finish(final)


    def checkRestoreWithMultipleObjectsInUndoRedo(self):
        # pylint:disable=too-many-statements
        from ZODB.FileStorage import FileStorage

        # Undo creates backpointers in (at least) FileStorage.  ZODB 3.2.1
        # FileStorage._data_find() had an off-by-8 error, neglecting to
        # account for the size of the backpointer when searching a
        # transaction with multiple data records.  The results were
        # unpredictable.  For example, it could raise a Python exception
        # due to passing a negative offset to file.seek(), or could
        # claim that a transaction didn't have data for an oid despite
        # that it actually did.
        #
        # The former failure mode was seen in real life, in a ZRS secondary
        # doing recovery.  On my box today, the second failure mode is
        # what happens in this test (with an unpatched _data_find, of
        # course).  Note that the error can only "bite" if more than one
        # data record is in a transaction, and the oid we're looking for
        # follows at least one data record with a backpointer.
        #
        # Unfortunately, _data_find() is a low-level implementation detail,
        # and this test does some horrid white-box abuse to test it.

        is_filestorage = isinstance(self._storage, FileStorage)
        closing = self._closing
        db = closing(DB(self._storage))
        c = closing(db.open())
        r = c.root()

        # Create some objects.
        r["obj1"] = MinPO(1)
        r["obj2"] = MinPO(1)
        transaction.commit()

        # Add x attributes to them.
        r["obj1"].x = 'x1'
        r["obj2"].x = 'x2'
        transaction.commit()

        c2 = closing(db.open())

        r = c2.root()
        self.assertEqual(r["obj1"].x, 'x1')
        self.assertEqual(r["obj2"].x, 'x2')

        # Dirty tricks.
        if is_filestorage:
            obj1_oid = r["obj1"]._p_oid
            obj2_oid = r["obj2"]._p_oid
            # This will be the offset of the next transaction, which
            # will contain two backpointers.
            pos = self._storage.getSize()

        # Undo the attribute creation.
        info = self._storage.undoInfo()
        tid = info[0]['id']
        t = Transaction()
        self._storage.tpc_begin(t)
        _oids = self._storage.undo(tid, t)
        self._storage.tpc_vote(t)
        self._storage.tpc_finish(t)

        r = closing(db.open()).root()
        self.assertRaises(AttributeError, getattr, r["obj1"], 'x')
        self.assertRaises(AttributeError, getattr, r["obj2"], 'x')

        if is_filestorage:
            # _data_find should find data records for both objects in that
            # transaction.  Without the patch, the second assert failed
            # (it claimed it couldn't find a data record for obj2) on my
            # box, but other failure modes were possible.
            self.assertTrue(self._storage._data_find(pos, obj1_oid, '') > 0)
            self.assertTrue(self._storage._data_find(pos, obj2_oid, '') > 0)

            # The offset of the next ("redo") transaction.
            pos = self._storage.getSize()

        # Undo the undo (restore the attributes).
        info = self._storage.undoInfo()
        tid = info[0]['id']
        t = Transaction()
        self._storage.tpc_begin(t)
        _oids = self._storage.undo(tid, t)
        self._storage.tpc_vote(t)
        self._storage.tpc_finish(t)

        c3 = closing(db.open())
        r = c3.root()
        self.assertEqual(r["obj1"].x, 'x1')
        self.assertEqual(r["obj2"].x, 'x2')

        if is_filestorage:
            # Again _data_find should find both objects in this txn, and
            # again the second assert failed on my box.
            self.assertTrue(self._storage._data_find(pos, obj1_oid, '') > 0)
            self.assertTrue(self._storage._data_find(pos, obj2_oid, '') > 0)

        # Indirectly provoke .restore().  .restore in turn indirectly
        # provokes _data_find too, but not usefully for the purposes of
        # the specific bug this test aims at:  copyTransactionsFrom() uses
        # storage iterators that chase backpointers themselves, and
        # return the data they point at instead.  The result is that
        # _data_find didn't actually see anything dangerous in this
        # part of the test.
        self._dst.copyTransactionsFrom(self._storage)
        self.compare(self._storage, self._dst)



# -*- coding: utf-8 -*-
"""
Tests for connections.py

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from hamcrest import assert_that
from nti.testing.matchers import validly_provides

from relstorage.tests import TestCase
from relstorage.tests import MockConnectionManager

from ..interfaces import IManagedStoreConnection
from ..interfaces import IManagedLoadConnection
from ..interfaces import IManagedDBConnection
from ..connections import StoreConnection
from ..connections import LoadConnection
from ..connections import ClosedConnection

class TestConnectionCommon(TestCase):

    klass = StoreConnection
    iface = IManagedStoreConnection

    def _makeArgument(self):
        return MockConnectionManager()

    def _makeOne(self):
        return self.klass(self._makeArgument())

    def test_provides(self):
        assert_that(self._makeOne(), validly_provides(self.iface))

class TestConnection(TestConnectionCommon):

    def test_restart_and_call_does_not_activate(self):
        manager = self._makeOne()
        self.assertFalse(manager.active)
        self.assertFalse(manager)

        def on_first_use(_conn, _cursor):
            self.fail('Should not call')

        cc = []

        def on_opened(conn, cursor):
            self.assertIsNotNone(conn)
            self.assertIsNotNone(cursor)
            cc.append(conn)
            cc.append(cursor)

        manager.on_first_use = on_first_use
        manager.on_opened = on_opened

        manager.restart_and_call(lambda _conn, _cursor: None)

        # If we try to access the cursor, it gets called, though.
        with self.assertRaises(AssertionError):
            getattr(manager, 'cursor')

        del manager.on_opened
        del manager.on_first_use
        self.assertEqual(cc, [manager.connection, manager.cursor])

    def test_restart_and_call_opens(self):
        manager = self._makeOne()
        opened = [0]
        def on_opened(_conn, _cursor):
            opened[0] += 1
        manager.on_opened = on_opened
        manager.restart_and_call(lambda _conn, _cursor: None)

        self.assertEqual(opened, [1])

    def test_restart_and_call_rollback(self):
        manager = self._makeOne()
        opened = [0]
        def on_rolledback(_conn, _cursor):
            opened[0] += 1
        manager.on_rolledback = on_rolledback

        manager.restart_and_call(lambda _conn, _cursor: None)
        # That was just an open, so no rollback
        self.assertEqual(opened, [0])

        # But second time is the charm
        manager.restart_and_call(lambda _conn, _cursor: None)
        self.assertEqual(opened, [1])

    def test_drop_activates(self, meth='drop'):
        count = [0]

        def on_first_use(_conn, _cursor):
            count[0] += 1

        manager = self._makeOne()
        manager.on_first_use = on_first_use

        getattr(manager, 'cursor')
        self.assertEqual(count, [1])

        getattr(manager, meth)()
        self.assertEqual(count, [1])

        getattr(manager, 'cursor')
        self.assertEqual(count, [2])

    def test_rollback_activates(self):
        self.test_drop_activates('rollback_quietly')

    def test_call_reconnect(self):
        manager = self._makeOne()
        manager.open_if_needed()
        called = []
        def f(_conn, _cur, fresh):
            called.append(fresh)
            if not fresh:
                raise manager.connmanager.driver.disconnected_exceptions[0]

        manager.call(f, can_reconnect=True)

        self.assertEqual(called, [False, True])

class TestLoadConnection(TestConnection):
    klass = LoadConnection
    iface = IManagedLoadConnection

class TestClosedConnection(TestConnectionCommon):
    klass = ClosedConnection
    iface = IManagedDBConnection



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

"""
PostgreSQL IDBDriverOptions implementation.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from ...drivers import implement_db_driver_options
from ...drivers import AbstractModuleDriver
from ...sql import DefaultDialect

logger = __import__('logging').getLogger(__name__)

class PostgreSQLDialect(DefaultDialect):
    """
    The defaults are setup for PostgreSQL.
    """

class AbstractPostgreSQLDriver(AbstractModuleDriver):
    dialect = PostgreSQLDialect()
    # Can we bundle statements into a single string?
    # "SELECT 1; COMMIT;"
    supports_multiple_statement_execute = True

    # Can we use the COPY command (copy_export)?
    supports_copy = True

    # PostgreSQL is the database most likey to generate
    # server-sent messages. Log those using a logger that
    # includes that name.
    message_logger = logger

    def connect_with_isolation(self, dsn,
                               isolation=None,
                               read_only=False,
                               deferrable=False,
                               application_name=None):
        """
        Return a connection whose transactions will have the defined
        characteristics and which will use the given
        ``application_name``.

        The driver should pass the ``application_name`` as part of the
        connect packet, and *not* via ``SET SESSION application_name``.
        This is because the former is both faster and
        because the latter does not work when connecting to a hot
        standby (it triggers "cannot set transaction read-write mode
        during recovery"). If it cannot be passed at connect time,
        an alternative is ``SELECT set_config('application_name', %s, FALSE)``
        which does not seem to have this problem.
        """
        raise NotImplementedError

    def set_lock_timeout(self, cursor, timeout):
        # PG8000 needs a literal embedded in the string; prepared
        # statements can't be SET with a variable.
        cursor.execute('SET lock_timeout = %s', (timeout,))

    def get_messages(self, conn):
        notices = conn.notices or ()
        if notices:
            notices = list(notices)
            if isinstance(notices[0], dict):
                # pg8000
                notices = [d[b'M'] for d in notices]
                conn.notices.clear()
            else:
                notices = list(notices)
                del conn.notices[:]
        return notices


    def synchronize_cursor_for_rollback(self, cursor):
        """Does nothing."""

    def execute_multiple_statement_with_hidden_commit(self, conn, cursor, stmt, params):
        # Exit the critical phase now. We don't have a fine-grained
        # way of doing this between statements, so up front is the fastest we
        # can exit. That's fine, Python isn't involved during releasing the locks; we're
        # either going to get the commit lock and successfully commit, or we're going to
        # fail. Either way, there's nothing we need to do to let other greenlets get going.
        # This is just like how MySQL handles it in ``callproc_multi_result``.
        self.exit_critical_phase(conn, cursor)
        cursor.execute(stmt, params)

database_type = 'postgresql'

implement_db_driver_options(
    __name__,
    'pg8000', 'psycopg2', 'psycopg2cffi',
)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest

from relstorage.cache import local_database
from relstorage.cache.local_database import Database
from relstorage.cache.persistence import sqlite_connect

from relstorage.tests import TestCase
from relstorage.cache.tests import MockOptions

class MockOptionsWithMemoryDB(MockOptions):
    cache_local_dir = ':memory:'


class UpdateTests(TestCase):
    # Tests to specifically cover the cases that
    # UPSERTS or multi-column updates aren't available.

    USE_UPSERT = False

    def setUp(self):
        self.options = MockOptionsWithMemoryDB()
        self.connection = sqlite_connect(
            self.options, "pfx-ignored")
        assert self.connection.rs_db_filename == ':memory:', self.connection
        self.db = self._makeOne()

    def tearDown(self):
        # Be sure we can commit; this can be an issue on PyPy
        self.connection.commit()
        self.db.close()

    def _makeOne(self):
        db = Database.from_connection(
            self.connection,
            use_upsert=self.USE_UPSERT,
        )
        ost = db.store_temp
        def store_temp(rows):
            return ost([
                (r[0], r[1], 0, r[2], r[3])
                for r
                in rows
            ])
        db.store_temp = store_temp
        db.real_store_temp = ost
        return db


    def test_set_checkpoints(self):
        self.db.update_checkpoints(1, 0)
        self.assertEqual(self.db.checkpoints, (1, 0))

    def test_update_checkpoints_newer(self):
        self.db.update_checkpoints(1, 0)
        self.db.update_checkpoints(2, 1)
        self.assertEqual(self.db.checkpoints, (2, 1))

    def test_update_checkpoints_older(self):
        self.db.update_checkpoints(2, 1)
        self.db.update_checkpoints(1, 0)
        self.assertEqual(self.db.checkpoints, (2, 1))

    def test_move_from_temp_empty(self):
        rows = [
            (0, 0, b'', 0),
            (1, 0, b'', 0)
        ]
        self.db.store_temp(rows)
        self.db.move_from_temp()
        self.assertEqual(dict(self.db.oid_to_tid),
                         {0: 0, 1: 0})

    def test_move_from_temp_mixed_updates(self):
        rows = [
            (0, 1, 0, b'0', 0),
            (1, 1, 0, b'1', 0),
            (2, 1, 0, b'2', 0),
        ]
        self.db.real_store_temp(rows)
        self.db.move_from_temp()

        new_rows = [
            # 0 goes backwards
            (0, 0, 0, b'-1', 0),
            # 1 stays the same (but we use a different state
            # to verify)
            (1, 1, 0, b'-1', 0),
            # 2 moves forward
            (2, 2, 0, b'2b', 6)
        ]

        self.db.real_store_temp(new_rows)
        self.db.move_from_temp()

        self.connection.commit()
        self.assertEqual(
            dict(self.db.oid_to_tid),
            {0: 1, 1: 1, 2: 2}
        )

        rows_in_db = list(self.db.fetch_rows_by_priority())
        rows_in_db.sort()
        self.assertEqual(rows_in_db[0], (0, 1, b'0', 1, 0))
        self.assertEqual(rows_in_db[1], (1, 1, b'-1', 1, 0))
        self.assertEqual(rows_in_db[2], (2, 2, b'2b', 2, 6))

    def test_remove_invalid_persistent_oids(self):
        rows = [
            (0, 1, b'0', 0),
            (1, 1, b'0', 0),
        ]

        self.db.store_temp(rows)
        self.db.move_from_temp()

        invalid_oids = range(1, 5000)
        count = self.db.remove_invalid_persistent_oids(invalid_oids)
        self.assertEqual(dict(self.db.oid_to_tid), {0: 1})
        self.assertEqual(count, len(invalid_oids))


    def test_trim_to_size_deletes_stale(self):
        rows = [
            (0, 1, b'0', 0),
            (1, 1, b'0', 0),
        ]
        self.db.store_temp(rows)
        self.db.move_from_temp()

        self.db.trim_to_size(
            # It's not trimming for a limit
            10000,
            # But we do know stale things that have to go.
            # OID 1 must be at least TID 2
            {1: 2}
        )
        # Leaving behind only one row
        self.assertEqual(dict(self.db.oid_to_tid), {0: 1})

    def test_trim_to_size_removes_size_oldest_first(self):
        # We delete the oldest, least used, biggest objects first.
        # These are tied on frequency, size, and transaction age.
        # Tie-breaker is OID, which indicates an older object again.
        rows = [
            (0, 1, b'0', 0),
            (1, 1, b'0', 0),
        ]
        self.db.store_temp(rows)
        self.db.move_from_temp()

        self.db.trim_to_size(
            1,
            ()
        )
        self.assertEqual(dict(self.db.oid_to_tid), {1: 1})

    def test_trim_to_size_removes_least_frequent_first(self):
        # We delete the oldest, least used, biggest objects first.
        # The newer object is less frequent than the older, so it goes.
        rows = [
            (0, 1, b'0', 1),
            (1, 1, b'0', 0),
        ]
        self.db.store_temp(rows)
        self.db.move_from_temp()

        self.db.trim_to_size(
            1,
            ()
        )
        self.assertEqual(dict(self.db.oid_to_tid), {0: 1})

    def test_trim_to_size_removes_biggest_first(self):
        # We delete the oldest, least used, biggest objects first.
        # The newer object is bigger, so it goes
        rows = [
            (0, 1, b'0', 0),
            (1, 1, b'00', 0),
        ]
        self.db.store_temp(rows)
        self.db.move_from_temp()

        self.db.trim_to_size(
            1,
            ()
        )
        self.assertEqual(dict(self.db.oid_to_tid), {0: 1})

    def test_total_state_len_with_embedded_nulls(self):
        # automatic text-infinity breaks LONGETH()
        # when there are embedded nul characters.
        # https://github.com/zodb/relstorage/issues/317
        #
        # In case the differences depend on the type of the parameter,
        # (they don't, not exactly)
        # we explicitly test bytes, unicode, and native
        # strings. What they do depend on is the version of Python.
        # In Python 2, the type and length of the five rows are
        #     text(1), text(1), text(1), text(3), text(3)
        # In Python 3, the type and length are
        #     blob(3), text(1), text(1), text(3), blob(3)
        rows = [
            (0, 1, b'1\x003', 0),
            (1, 1, u'2\x003', 0),
            (2, 2, '3\x003', 0),
            (3, 3, '303', 0),
            (4, 3, b'403', 0),
        ]
        self.db.store_temp(rows)
        self.db.move_from_temp()

        self.assertEqual(sum(len(row[2]) for row in rows),
                         15)

        self.assertEqual(self.db.total_state_len,
                         sum(len(row[2]) for row in rows))

@unittest.skipIf(not local_database.SUPPORTS_UPSERT,
                 "Requires upserts")
class UpsertUpdateTests(UpdateTests):
    USE_UPSERT = True


class MultiConnectionTests(TestCase):

    timeout = 0

    def setUp(self):
        import tempfile
        import shutil
        self.options = MockOptionsWithMemoryDB()
        tempdir = tempfile.mkdtemp('.rstest')
        self.addCleanup(shutil.rmtree, tempdir, True)
        self.options.cache_local_dir = tempdir
        self.connection = self.connect()
        self.addCleanup(self.connection.close)
        self.db = self._makeOne()
        self.addCleanup(self.db.close)

    def connect(self):
        return sqlite_connect(
            self.options, "pfx-ignored",
            timeout=self.timeout
        )

    def tearDown(self):
        # Be sure we can commit; this can be an issue on PyPy
        self.connection.commit()
        self.db.close()

    def _makeOne(self, conn=None):
        return Database.from_connection(
            conn or self.connection,
        )

    def test_delete_oids_other_open_transaction(self):
        rows = [
            (0, 1, 0, b'0', 0),
            (1, 1, 0, b'0', 0),
        ]
        self.db.store_temp(rows)
        self.db.move_from_temp()

        invalid_oids = range(1, 5000)

        conn2 = self.connect()
        try:
            cur2 = conn2.cursor()
            cur2.execute("BEGIN")
            cur2.execute('DELETE FROM object_state WHERE zoid = 0')

            count = self.db.remove_invalid_persistent_oids(invalid_oids)
            self.assertEqual(count, -1)
            cur2.close()
        finally:
            conn2.close()



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""IObjectMover implementation.
"""
from __future__ import absolute_import
from __future__ import print_function

import os
from hashlib import md5

from zope.interface import implementer

from .._compat import OID_TID_MAP_TYPE
from .._compat import metricmethod_sampled
from .._compat import metricmethod
from ._util import noop_when_history_free
from ._util import query_property as _query_property
from ._util import DatabaseHelpersMixin
from .._compat import ABC
from .batch import RowBatcher
from .interfaces import IObjectMover
from .interfaces import AggregateOperationTimeoutError
from .schema import Schema
from .sql import it
from .sql.schema import ColumnExpression

objects = Schema.all_current_object_state
object_state = Schema.object_state


def _compute_md5sum(_self, data):
    if data is None:
        return None
    return md5(data).hexdigest()



@implementer(IObjectMover)
class AbstractObjectMover(DatabaseHelpersMixin, ABC):

    def __init__(self, database_driver, options, runner=None,
                 version_detector=None,
                 batcher_factory=RowBatcher):
        """
        :param database_driver: The `IDBDriver` in use.
        """
        self.driver = database_driver
        self.keep_history = options.keep_history
        self.blob_chunk_size = options.blob_chunk_size
        self.runner = runner

        self.version_detector = version_detector
        self.make_batcher = batcher_factory

    _compute_md5sum = noop_when_history_free(_compute_md5sum)

    _load_current_query = objects.select(
        objects.c.state, objects.c.tid
    ).where(
        objects.c.zoid == objects.orderedbindparam()
    ).prepared()

    @metricmethod_sampled
    def load_current(self, cursor, oid):
        """Returns the current pickle and integer tid for an object.

        oid is an integer.  Returns (None, None) if object does not exist.
        """
        stmt = self._load_current_query
        stmt.execute(cursor, (oid,))
        # Note that we cannot rely on cursor.rowcount being
        # a valid indicator. The DB-API doesn't require it, and
        # some implementations, like MySQL Connector/Python are
        # unbuffered by default and can't provide it.
        row = cursor.fetchone()
        if row:
            state, tid = row
            state = self.driver.binary_column_as_state_type(state)
            # If it's None, the object's creation has been
            # undone.
            return state, tid

        return None, None

    _load_currents_queries = (
        (('zoid', 'state', 'tid'), 'current_object JOIN object_state USING(zoid, tid)', 'zoid'),
        (('zoid', 'state', 'tid'), 'object_state', 'zoid'),
    )

    _load_currents_query = _query_property('_load_currents')

    @metricmethod_sampled
    def load_currents(self, cursor, oids):
        """Returns the current (oid, state, tid) for specified object ids."""
        columns, table, filter_column = self._load_currents_query
        binary_column_as_state_type = self.driver.binary_column_as_state_type
        batcher = self.make_batcher(cursor, row_limit=1000)
        rows = batcher.select_from(columns, table, **{filter_column: oids})
        for row in rows:
            oid, state, tid = row
            yield oid, binary_column_as_state_type(state), tid

    _load_revision_query = object_state.select(
        object_state.c.state
    ).where(
        object_state.c.zoid == object_state.orderedbindparam()
    ).and_(
        object_state.c.tid == object_state.orderedbindparam()
    ).prepared()

    @metricmethod_sampled
    def load_revision(self, cursor, oid, tid):
        """Returns the pickle for an object on a particular transaction.

        Returns None if no such state exists.
        """
        stmt = self._load_revision_query
        stmt.execute(cursor, (oid, tid))
        row = cursor.fetchone()
        if row:
            (state,) = row
            return self.driver.binary_column_as_state_type(state)
        return None

    _exists_query = Schema.all_current_object.select(
        Schema.all_current_object.c.zoid
    ).where(
        Schema.all_current_object.c.zoid == Schema.all_current_object.orderedbindparam()
    )

    @metricmethod_sampled
    def exists(self, cursor, oid):
        """Returns a true value if the given object exists."""
        stmt = self._exists_query
        stmt.execute(cursor, (oid,))
        row = cursor.fetchone()
        return row

    _load_before_query = object_state.select(
        object_state.c.state, object_state.c.tid
    ).where(
        object_state.c.zoid == object_state.orderedbindparam()
    ).and_(
        object_state.c.tid < object_state.orderedbindparam()
    ).order_by(
        object_state.c.tid, "DESC"
    ).limit(
        1
    ).prepared()


    @metricmethod_sampled
    def load_before(self, cursor, oid, tid):
        """Returns the pickle and tid of an object before transaction tid.

        Returns (None, None) if no earlier state exists.
        """

        self._load_before_query.execute(cursor, (oid, tid))
        row = cursor.fetchone()
        if row:
            state, tid = row
            state = self.driver.binary_column_as_state_type(state)
            # None in state means The object's creation has been undone
            return state, tid
        return None, None

    _get_tid_after_query = object_state.select(
        object_state.c.tid
    ).where(
        object_state.c.zoid == object_state.orderedbindparam()
    ).and_(
        object_state.c.tid > object_state.orderedbindparam()
    ).order_by(
        object_state.c.tid, "ASC"
    ).limit(
        1
    ).prepared()


    @metricmethod_sampled
    def get_object_tid_after(self, cursor, oid, tid):
        """Returns the tid of the next change after an object revision.

        Returns None if no later state exists.
        """
        self._get_tid_after_query.execute(cursor, (oid, tid))
        row = cursor.fetchone()
        if row:
            return row[0]

    _current_object_tids_queries = (
        (('zoid', 'tid'), 'current_object', 'zoid'),
        (('zoid', 'tid'), 'object_state', 'zoid'),
    )

    _current_object_tids_query = _query_property('_current_object_tids')

    _current_object_tids_map_type = OID_TID_MAP_TYPE

    @metricmethod
    def current_object_tids(self, cursor, oids, timeout=None):
        """Returns the current {oid: tid} for specified object ids."""
        # This is a metricmethod, not a metricmethod_sampled because only databases that
        # use composed lock_objects_and_detect_conflicts call it for every transaction,
        # and even then they only call it once, so its like the tpc_* methods.
        # Other databases use it only when restoring the cache at startup (just once) so its
        # unlikely to get sampled (see relstorage.cache.mvcc).
        res = self._current_object_tids_map_type()
        columns, table, filter_column = self._current_object_tids_query
        batcher = self.make_batcher(cursor)
        rows = batcher.select_from(columns, table, timeout=timeout, **{filter_column: oids})
        if timeout:
            # Do the collecting and iterating in Python so we can handle partial results
            res = self._current_object_tids_map_type()
            try:
                for (oid, tid) in rows:
                    res[oid] = tid
            except AggregateOperationTimeoutError as ex:
                ex.partial_result = res
                raise
        else:
            # Do the collecting and iterating in C
            res = self._current_object_tids_map_type(list(rows))
        return res


    def on_store_opened(self, cursor, restart=False):
        """
        Hook for subclasses.
        """

    def on_load_opened(self, cursor, restart=False):
        """
        Hook for subclasses.
        """

    _store_temp_query = Schema.temp_store.upsert(
        it.c.zoid,
        it.c.prev_tid,
        it.c.md5,
        it.c.state
    ).on_conflict(
        it.c.zoid
    ).do_update(
        it.c.prev_tid,
        it.c.md5,
        it.c.state
    )

    @metricmethod_sampled
    def store_temps(self, cursor, state_oid_tid_iter):
        """
        Uses the cursor's ``executemany`` method to store temporary
        objects.

        If there is a more optimal way to implement putting objects in
        the database, please do so.

            - On SQLite, ``executemany`` is implemnted in a C looping
              over the provided iterator. Which it turns out is
              exactly what the normal ``execute`` method also does (it
              just uses a one-row iterator). So ``executemany`` that
              saves substantial setup overhead dealing with sqlite's
              prepared statements.

            - On Postgresql, we use COPY for this (unless we're using
              the 'gevent psycopg2' driver; it's the only thing that
              doesn't support COPY). None of the supported PostgreSQL
              drivers have a good ``executemany`` method, so they
              should fall back to using our own RowBatcher.

            - On Oracle, we use the RowBatcher with a combination of
              bulk array operations and direct inserts.

            - On MySQL, the preferred driver (mysqlclient) has a
              decent implementation of executemany for INSERT or
              REPLACE (basically an optimized form of what our
              RowBatcher does). That implementation is shared with
              PyMySQL as well, but it must be a simple INSERT
              statement matching a regular expression. Note that it
              has a bug though: it can't handle an iterator that's
              empty.
        """
        query = self._store_temp_query
        do_md5 = self._compute_md5sum
        Binary = self.driver.Binary

        query.executemany(
            cursor,
            (
                (oid_int, tid_int, do_md5(data), Binary(data))
                for (data, oid_int, tid_int)
                in state_oid_tid_iter
            )
        )

    @metricmethod_sampled
    def replace_temps(self, cursor, state_oid_tid_iter):
        """
        Assumes that ``store_temps`` is using an upsert query and simply calls
        that method.

        The same comments apply. In particular,
        MySQLclient won't optimize an UPDATE in the same way it does an INSERT.
        """
        self.store_temps(cursor, state_oid_tid_iter)

    @metricmethod_sampled
    def _generic_restore(self, batcher, oid, tid, data,
                         command, suffix):
        """
        Store an object directly, without conflict detection.

        Used for copying transactions into this database.

        Either the *command* or the *suffix* must be capable of
        handling conflicts in a single query. For example,
        ``command='INSERT OR REPLACE'``
        or ``command='INSERT', suffix='ON CONFLICT (zoid) DO...``
        """
        if data is not None:
            encoded = self.driver.Binary(data)
            size = len(data)
        else:
            encoded = None
            size = 0

        if self.keep_history:
            # We can record deletion/un-creation via a null state.
            md5sum = self._compute_md5sum(data)
            row_schema = """
                %s, %s,
                COALESCE((SELECT tid FROM current_object WHERE zoid = %s), 0),
                %s, %s, %s
            """
            batcher.insert_into(
                "object_state (zoid, tid, prev_tid, md5, state_size, state)",
                row_schema,
                (oid, tid, oid, md5sum, size, encoded),
                rowkey=(oid, tid),
                size=size,
                command=command,
                suffix=suffix
            )
        else:
            # history free can only delete the entire record.
            if data:
                batcher.insert_into(
                    "object_state (zoid, tid, state_size, state)",
                    "%s, %s, %s, %s",
                    (oid, tid, size, encoded),
                    rowkey=oid,
                    size=size,
                    command=command,
                    suffix=suffix
                )
            else:
                batcher.delete_from('object_state', zoid=oid)

    def restore(self, cursor, batcher, oid, tid, data):
        raise NotImplementedError()

    _detect_conflict_query = Schema.temp_store.inner_join(
        Schema.all_current_object_state
    ).using(
        Schema.all_current_object_state.c.zoid
    ).select(
        Schema.temp_store.c.zoid,
        Schema.all_current_object_state.c.tid,
        Schema.temp_store.c.prev_tid,
        Schema.all_current_object_state.c.state,
    ).where(
        # Some databases may be able to benefit from prev_tid <> 0, but it depends
        # on their ability to make use of indexes
        Schema.all_current_object_state.c.tid != Schema.temp_store.c.prev_tid
    ).prepared()

    @metricmethod_sampled
    def detect_conflict(self, cursor):
        self._detect_conflict_query.execute(cursor)
        # Note that we're not transforming the state into
        # bytes; it doesn't seem to be needed here, even with sqlite3
        # on Python 2 (where it is a buffer).
        rows = cursor.fetchall()
        return rows

    # Subclasses may override any of these queries if there is a
    # more optimal form.
    _move_from_temp_hp_insert_query = Schema.object_state.insert(
    ).from_select(
        (Schema.object_state.c.zoid,
         Schema.object_state.c.tid,
         Schema.object_state.c.prev_tid,
         Schema.object_state.c.md5,
         Schema.object_state.c.state_size,
         Schema.object_state.c.state),
        Schema.temp_store.select(
            Schema.temp_store.c.zoid,
            Schema.temp_store.orderedbindparam(),
            Schema.temp_store.c.prev_tid,
            Schema.temp_store.c.md5,
            'COALESCE(LENGTH(state), 0)',
            Schema.temp_store.c.state
        ).order_by(
            Schema.temp_store.c.zoid
        )
    ).prepared()

    _move_from_temp_hf_upsert_query = Schema.object_state.upsert(
        it.c.zoid,
        it.c.tid,
        it.c.state_size,
        it.c.state
    ).from_select(
        (Schema.temp_store.c.zoid,
         Schema.object_state.c.tid, # correct column for typing
         Schema.object_state.c.state_size,
         Schema.temp_store.c.state),
        Schema.temp_store.select(
            it.c.zoid,
            ColumnExpression(it.orderedbindparam()).aliased('tid'),
            ColumnExpression('COALESCE(LENGTH(state), 0)').aliased('state_size'),
            it.c.state
        ).order_by(
            it.c.zoid
        )
    ).on_conflict(
        it.c.zoid
    ).do_update(
        it.c.state,
        it.c.tid,
        it.c.state_size
    ).prepared()

    _move_from_temp_copy_blob_query = Schema.blob_chunk.insert(
    ).from_select(
        (Schema.blob_chunk.c.zoid,
         Schema.blob_chunk.c.tid,
         Schema.blob_chunk.c.chunk_num,
         Schema.blob_chunk.c.chunk),
        Schema.temp_blob_chunk.select(
            Schema.temp_blob_chunk.c.zoid,
            Schema.temp_blob_chunk.orderedbindparam(),
            Schema.temp_blob_chunk.c.chunk_num,
            Schema.temp_blob_chunk.c.chunk
        )
    ).prepared()


    _move_from_temp_hf_delete_blob_chunk_query = """
    DELETE FROM blob_chunk
    WHERE zoid IN (SELECT zoid FROM temp_store)
    """


    def _move_from_temp_object_state(self, cursor, tid):
        """
        Called for history-free databases.

        Should replace all entries in object_state with the same zoid
        from temp_store.

        This implementation is in two steps, first deleting from
        ``object_state`` with :attr:`_move_from_temp_hf_delete_query`,
        and then copying from ``temp_store`` using
        :attr:`_move_from_temp_hf_insert_query`.

        If a subclass can do this in a single step with an ``UPSERT``,
        it should set :attr:`_move_from_temp_hf_delete_query` to a
        false value.

        Recall that the queries that touch ``current_object`` and
        ``object_state`` need to be certain the order they use (by
        ``zoid``) to avoid deadlocks.

        Blobs are handled separately.
        """
        stmt = self._move_from_temp_hf_upsert_query
        __traceback_info__ = stmt
        stmt.execute(cursor, (tid,))


    @metricmethod_sampled
    def move_from_temp(self, cursor, tid, txn_has_blobs):
        """
        Move the temporarily stored objects to permanent storage.
        """
        if self.keep_history:
            stmt = self._move_from_temp_hp_insert_query
            __traceback_info__ = stmt
            stmt.execute(cursor, (tid,))
        else:
            self._move_from_temp_object_state(cursor, tid)

            if txn_has_blobs:
                # If we can require storages to have an UPSERT (mysql and
                # postgres do), then can we remove the DELETE?
                # Answer: probably not. What if the blob shrunk and we
                # have fewer chunks than we used to?
                stmt = self._move_from_temp_hf_delete_blob_chunk_query
                cursor.execute(stmt)

        if txn_has_blobs:
            stmt = self._move_from_temp_copy_blob_query
            __traceabck_info__ = stmt
            stmt.execute(cursor, (tid,))


    # Insert and update current objects.
    # Note that to avoid deadlocks, it is incredibly important
    # to order the updates in OID order.
    _update_current_upsert_query = Schema.current_object.upsert(
        it.c.zoid, it.c.tid
    ).from_select(
        (it.c.zoid, it.c.tid),
        Schema.object_state.select(
            it.c.zoid, Schema.object_state.c.tid # correct column for typing
        ).where(
            it.c.tid == it.orderedbindparam()
        ).order_by(it.c.zoid)
    ).on_conflict(
        it.c.zoid
    ).do_update(
        it.c.tid
    ).prepared()

    @noop_when_history_free
    @metricmethod_sampled
    def update_current(self, cursor, tid):
        """
        Update the current object pointers.

        tid is the integer tid of the transaction being committed.
        """
        stmt = self._update_current_upsert_query
        stmt.execute(cursor, (tid,))

    @metricmethod_sampled
    def download_blob(self, cursor, oid, tid, filename):
        """Download a blob into a file."""
        stmt = """
        SELECT chunk
        FROM blob_chunk
        WHERE zoid = %s
            AND tid = %s
            AND chunk_num = %s
        """

        f = None
        bytecount = 0
        try:
            chunk_num = 0
            while True:
                cursor.execute(stmt, (oid, tid, chunk_num))
                rows = list(cursor)
                if rows:
                    assert len(rows) == 1
                    chunk = rows[0][0]
                else:
                    # No more chunks.  Note: if there are no chunks at
                    # all, then this method should not write a file.
                    break

                if f is None:
                    f = open(filename, 'wb')

                f.write(chunk)
                bytecount += len(chunk)
                chunk_num += 1
        except:
            if f is not None:
                f.close()
                os.remove(filename)
            raise

        if f is not None:
            f.close()
        return bytecount

    _upload_blob_uses_chunks = True

    def _upload_blob_clear_old_blob(self, cursor, oid, tid):
        if tid is not None:
            if self.keep_history:
                delete_stmt = """
                DELETE FROM blob_chunk
                WHERE zoid = %s AND tid = %s
                """
                cursor.execute(delete_stmt, (oid, tid))
            else:
                delete_stmt = "DELETE FROM blob_chunk WHERE zoid = %s"
                cursor.execute(delete_stmt, (oid,))

            use_tid = True
            insert_stmt = """
            INSERT INTO blob_chunk (zoid, tid, chunk_num, chunk)
            VALUES (%s, %s, %s, %s)
            """
        else:
            use_tid = False
            delete_stmt = "DELETE FROM temp_blob_chunk WHERE zoid = %s"
            cursor.execute(delete_stmt, (oid,))

            insert_stmt = """
            INSERT INTO temp_blob_chunk (zoid, chunk_num, chunk)
            VALUES (%s, %s, %s)
            """

        return insert_stmt, use_tid

    def _upload_blob_read_chunks(self, cursor, oid, tid, filename,
                                 use_chunks, insert_stmt, use_tid):
        Binary = self.driver.Binary
        with open(filename, 'rb') as f:
            chunk_num = 0
            while True:
                chunk = f.read(self.blob_chunk_size) if use_chunks else f.read()
                if not chunk and chunk_num > 0:
                    # EOF.  Note that we always write at least one
                    # chunk, even if the blob file is empty.
                    break
                if use_tid:
                    params = (oid, tid, chunk_num, Binary(chunk))
                else:
                    params = (oid, chunk_num, Binary(chunk))
                cursor.execute(insert_stmt, params)
                chunk_num += 1

    @metricmethod_sampled
    def upload_blob(self, cursor, oid, tid, filename):
        """Upload a blob from a file.

        If serial is None, upload to the temporary table.
        """
        insert_stmt, use_tid = self._upload_blob_clear_old_blob(cursor, oid, tid)
        self._upload_blob_read_chunks(
            cursor, oid, tid, filename,
            self._upload_blob_uses_chunks, insert_stmt, use_tid
        )


class RowBatcherStoreTemps(object):
    """
    A helper class to implement ``store_temps`` using a RowBatcher.
    You must provide an implementation of
    :meth:`store_temp_into_batcher` and it must be an upsert. The
    :meth:`generic_store_temp_into_batcher` method can be used to help
    with this.
    """

    def __init__(self, keep_history, binary, batcher_factory=RowBatcher):
        self.make_batcher = batcher_factory
        self.keep_history = keep_history
        self.binary = binary

    _compute_md5sum = noop_when_history_free(_compute_md5sum)

    @metricmethod_sampled
    def store_temps(self, cursor, state_oid_tid_iter):
        store_temp = self.store_temp_into_batcher
        batcher = self.make_batcher(cursor) # Default row limit
        for data, oid_int, tid_int in state_oid_tid_iter:
            store_temp(batcher, oid_int, tid_int, data)
        batcher.flush()

    replace_temps = store_temps

    # The _generic methods allow for UPSERTs, at least on MySQL
    # and PostgreSQL. Previously, MySQL used `command='REPLACE'`
    # for an UPSERT; now it uses a suffix 'ON DUPLICATE KEY UPDATE ...'.
    # PostgreSQL uses a suffix 'ON CONFLICT (...) UPDATE ...'.

    generic_command = 'INSERT'
    generic_suffix = ''

    def generic_store_temp_into_batcher(self, batcher, oid, prev_tid, data):
        md5sum = self._compute_md5sum(data)
        command = self.generic_command
        suffix = self.generic_suffix
        # TODO: Now that we guarantee not to feed duplicates here, drop
        # the conflict handling.
        if command == 'INSERT' and not suffix:
            batcher.delete_from('temp_store', zoid=oid)
        batcher.insert_into(
            "temp_store (zoid, prev_tid, md5, state)",
            batcher.row_schema_of_length(4),
            (oid, prev_tid, md5sum, self.binary(data)),
            rowkey=oid,
            size=len(data) + 32,
            command=command,
            suffix=suffix
        )

    store_temp_into_batcher = generic_store_temp_into_batcher



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2016 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
MySQLdb IDBDriver implementations.
"""
from __future__ import absolute_import
from __future__ import print_function

from zope.interface import implementer

from relstorage.adapters.interfaces import IDBDriver
from relstorage.adapters.interfaces import IDBDriverSupportsCritical
from relstorage.adapters.drivers import AbstractModuleDriver

from relstorage._util import Lazy

from . import AbstractMySQLDriver
from . import IterateFetchmanyMixin
from ...drivers import GeventDriverMixin

__all__ = [
    'MySQLdbDriver',
    'GeventMySQLdbDriver'
]

@implementer(IDBDriver)
class MySQLdbDriver(AbstractMySQLDriver):
    __name__ = 'MySQLdb'

    MODULE_NAME = 'MySQLdb'
    PRIORITY = 1
    PRIORITY_PYPY = 3
    _GEVENT_CAPABLE = False

    def synchronize_cursor_for_rollback(self, cursor):
        AbstractModuleDriver.synchronize_cursor_for_rollback(self, cursor)

    @Lazy
    def _server_side_cursor(self):
        from MySQLdb.cursors import SSCursor # pylint:disable=no-name-in-module,import-error
        class Cursor(IterateFetchmanyMixin, SSCursor):
            pass
        return Cursor

    @Lazy
    def _strict_cursor(self):
        SSCursor = self._server_side_cursor
        # Using MySQLdb.cursors.SSCursor can get us some legitimate
        # errors (ProgrammingError: (2014, "Commands out of sync; you
        # can't run this command now")), although it adds some overhead
        # because of more database communication. And the docstring says you have to
        # call `close()` before making another query, but in practice that
        # doesn't seem to be the case. You must consume everything though.
        #
        # For extra strictness/debugging, we can wrap this with our
        # custom debugging cursor.
        if 0: # pylint:disable=using-constant-test
            # TODO: Make this configurable, and add this to
            # all drivers.
            from relstorage._util import NeedsFetchallBeforeCloseCursor

            def cursor_factory(conn):
                cur = SSCursor(conn) # pylint:disable=too-many-function-args
                cur = NeedsFetchallBeforeCloseCursor(cur)
                return cur
        else:
            cursor_factory = SSCursor

        return cursor_factory

    def connect(self, *args, **kwargs):
        if self.STRICT and 'cursorclass' not in kwargs:
            kwargs['cursorclass'] = self._strict_cursor
        return AbstractMySQLDriver.connect(self, *args, **kwargs)

    def connection_may_need_rollback(self, conn):
        if conn.readonly:
            return True
        return True

    def connection_may_need_commit(self, conn):
        if conn.readonly:
            return False
        return True

@implementer(IDBDriverSupportsCritical)
class GeventMySQLdbDriver(GeventDriverMixin,
                          MySQLdbDriver):
    __name__ = 'gevent MySQLdb'

    _GEVENT_CAPABLE = True
    _GEVENT_NEEDS_SOCKET_PATCH = False

    # If we have more rows than this, it will take multiple trips to
    # the socket and C to read them. OTOH, that's a rough indication
    # of how often we will yield to the event loop. Using fetchall()
    # will yield between fetching this many rows, but all the results
    # will still be returned to the caller for processing in one
    # batch. Plain iteration will yield after this many rows, and return
    # rows to the caller one at a time.
    # cursor_arraysize = 1024 # We inherit cursor_arraysize from the abstract driver.

    def __init__(self):
        super(GeventMySQLdbDriver, self).__init__()
        # Replace self._connect (which was MySQLdb.connect) with
        # direct call to our desired class.
        self._connect = self._get_connection_class()
        self._strict_cursor = self._connect.default_cursor
        self._server_side_cursor = self._connect.default_cursor

    _Connection = None

    def enter_critical_phase_until_transaction_end(self, connection, cursor):
        # pylint:disable=unused-argument
        connection.enter_critical_phase_until_transaction_end()

    def is_in_critical_phase(self, connection, cursor):
        return connection.is_in_critical_phase()

    def exit_critical_phase(self, connection, cursor):
        connection.exit_critical_phase()

    @classmethod
    def _get_connection_class(cls):
        from ._mysqldb_gevent import Connection
        return Connection



# -*- coding: utf-8 -*-
"""
Tests for _mvcc.py

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from hamcrest import assert_that
from nti.testing.matchers import validly_provides

from .. import interfaces
from .. import _mvcc as mvcc
from . import TestCase

MockViewer = mvcc.DetachableMVCCDatabaseViewer

class TestDetachableMVCCDatabaseCoordinator(TestCase):

    def _makeOne(self):
        return mvcc.DetachableMVCCDatabaseCoordinator()

    def test_ctor(self):
        c = self._makeOne()
        self.assertEqual(c.minimum_highest_visible_tid, None)
        self.assertEqual(c.maximum_highest_visible_tid, None)

    def test_implements(self):
        assert_that(self._makeOne(),
                    validly_provides(interfaces.IMVCCDatabaseCoordinator))

    def test_register_no_hvt(self):
        c = self._makeOne()
        viewer = MockViewer()
        c.register(viewer)
        self.assertTrue(c.is_registered(viewer))
        self.assertEqual(0, c._viewer_count_at_min)
        self.assertEqual(None, c.maximum_highest_visible_tid)
        self.assertEqual(None, c.minimum_highest_visible_tid)

        viewer2 = MockViewer()
        c.register(viewer2)
        self.assertTrue(c.is_registered(viewer2))
        self.assertEqual(0, c._viewer_count_at_min)
        self.assertEqual(None, c.maximum_highest_visible_tid)
        self.assertEqual(None, c.minimum_highest_visible_tid)

        c.unregister(viewer)
        self.assertFalse(c.is_registered(viewer))
        self.assertEqual(0, c._viewer_count_at_min)
        self.assertEqual(None, c.maximum_highest_visible_tid)
        self.assertEqual(None, c.minimum_highest_visible_tid)

    def test_register_min_hvt(self):
        c = self._makeOne()
        viewer = MockViewer()
        viewer.highest_visible_tid = 1
        c.register(viewer)
        self.assertTrue(c.is_registered(viewer))
        self.assertEqual(1, c._viewer_count_at_min)
        self.assertEqual(1, c.maximum_highest_visible_tid)
        self.assertEqual(1, c.minimum_highest_visible_tid)

        # Unregistering someone not registered does nothing.
        viewer2 = MockViewer()
        viewer2.highest_visible_tid = 1
        c.unregister(viewer2)
        self.assertEqual(1, c._viewer_count_at_min)
        self.assertEqual(1, c.maximum_highest_visible_tid)
        self.assertEqual(1, c.minimum_highest_visible_tid)

        # dup registration does nothing
        viewer.highest_visible_tid = 2 # cheat
        c.register(viewer)
        self.assertTrue(c.is_registered(viewer))
        self.assertEqual(1, c._viewer_count_at_min)
        self.assertEqual(1, c.maximum_highest_visible_tid)
        self.assertEqual(1, c.minimum_highest_visible_tid)

        viewer.detached = False
        viewer.highest_visible_tid = 1
        c.unregister(viewer)
        self.assertFalse(c.is_registered(viewer))
        self.assertEqual(0, c._viewer_count_at_min)
        self.assertEqual(None, c.maximum_highest_visible_tid)
        self.assertEqual(None, c.minimum_highest_visible_tid)

    def test_register_multiple(self):
        c = self._makeOne()
        viewer = MockViewer()
        viewer.highest_visible_tid = 1
        c.register(viewer)
        self.assertTrue(c.is_registered(viewer))
        self.assertEqual(1, c._viewer_count_at_min)
        self.assertEqual(1, c.maximum_highest_visible_tid)
        self.assertEqual(1, c.minimum_highest_visible_tid)

        # Without setting a min hvt, does nothing
        viewer2 = MockViewer()
        c.register(viewer2)
        self.assertTrue(c.is_registered(viewer2))
        self.assertEqual(1, c._viewer_count_at_min)
        self.assertEqual(1, c.maximum_highest_visible_tid)
        self.assertEqual(1, c.minimum_highest_visible_tid)

    def test_register_multiple_min_hvt(self):
        c = self._makeOne()
        viewer = MockViewer()
        viewer.highest_visible_tid = 1
        c.register(viewer)
        self.assertTrue(c.is_registered(viewer))
        self.assertEqual(1, c._viewer_count_at_min)
        self.assertEqual(1, c.maximum_highest_visible_tid)
        self.assertEqual(1, c.minimum_highest_visible_tid)

        viewer2 = MockViewer()
        viewer2.highest_visible_tid = 1
        c.register(viewer2)
        self.assertTrue(c.is_registered(viewer2))
        self.assertEqual(2, c._viewer_count_at_min)
        self.assertEqual(1, c.maximum_highest_visible_tid)
        self.assertEqual(1, c.minimum_highest_visible_tid)

        c.unregister(viewer2)
        self.assertEqual(1, c._viewer_count_at_min)
        self.assertEqual(1, c.maximum_highest_visible_tid)
        self.assertEqual(1, c.minimum_highest_visible_tid)

    def test_detach(self):
        c = self._makeOne()
        viewer = MockViewer()
        viewer.highest_visible_tid = 1
        c.register(viewer)
        self.assertEqual(1, c._viewer_count_at_min)
        self.assertEqual(1, c.maximum_highest_visible_tid)
        self.assertEqual(1, c.minimum_highest_visible_tid)

        viewer1 = MockViewer()
        viewer1.highest_visible_tid = 1
        c.register(viewer1)
        self.assertEqual(2, c._viewer_count_at_min)
        self.assertEqual(1, c.maximum_highest_visible_tid)
        self.assertEqual(1, c.minimum_highest_visible_tid)

        viewer2 = MockViewer()
        viewer2.highest_visible_tid = 2
        c.register(viewer2)
        self.assertTrue(c.is_registered(viewer2))
        self.assertEqual(2, c._viewer_count_at_min)
        self.assertEqual(2, c.maximum_highest_visible_tid)
        self.assertEqual(1, c.minimum_highest_visible_tid)

        # Detach one of the guys making our minimum
        c.detach(viewer)
        self.assertEqual(1, c._viewer_count_at_min)
        self.assertEqual(2, c.maximum_highest_visible_tid)
        self.assertEqual(1, c.minimum_highest_visible_tid)

        # And again, no changes
        c.detach(viewer)
        self.assertEqual(1, c._viewer_count_at_min)
        self.assertEqual(2, c.maximum_highest_visible_tid)
        self.assertEqual(1, c.minimum_highest_visible_tid)

        # detach other minimum
        c.detach(viewer1)
        self.assertEqual(1, c._viewer_count_at_min)
        self.assertEqual(2, c.maximum_highest_visible_tid)
        self.assertEqual(2, c.minimum_highest_visible_tid)

        # detach the max
        c.detach(viewer2)
        self.assertEqual(0, c._viewer_count_at_min)
        self.assertEqual(None, c.maximum_highest_visible_tid)
        self.assertEqual(None, c.minimum_highest_visible_tid)

    def test_viewers_at_or_before(self):
        viewer1 = MockViewer()
        viewer1.highest_visible_tid = 1
        viewer2 = MockViewer()
        viewer2.highest_visible_tid = 2
        viewer3 = MockViewer()
        viewer3.highest_visible_tid = 3
        c = self._makeOne()

        c.register(viewer1)
        c.register(viewer2)
        c.register(viewer3)

        self.assertEqual(c.viewers_at_or_before(4), {viewer1, viewer2, viewer3})
        self.assertEqual(c.viewers_at_or_before(3), {viewer1, viewer2, viewer3})
        self.assertEqual(c.viewers_at_or_before(2), {viewer1, viewer2})
        self.assertEqual(c.viewers_at_or_before(1), {viewer1})
        self.assertEqual(c.viewers_at_minimum(), {viewer1})

        c.detach(viewer2)
        self.assertEqual(c.viewers_at_or_before(3), {viewer1, viewer3})
        self.assertEqual(c.viewers_at_or_before(2), {viewer1})

        c.change(viewer1, None)
        self.assertEqual(c.viewers_at_or_before(3), {viewer3})
        self.assertEqual(c.viewers_at_or_before(2), ())

        c.change(viewer3, 1)
        self.assertEqual(c.viewers_at_or_before(3), {viewer3})
        self.assertEqual(c.viewers_at_or_before(2), {viewer3})
        self.assertEqual(c.viewers_at_or_before(1), {viewer3})

        c.change(viewer1, 4)
        self.assertEqual(c.viewers_at_or_before(4), {viewer3, viewer1})
        self.assertEqual(c.viewers_at_or_before(3), {viewer3})
        self.assertEqual(c.viewers_at_or_before(2), {viewer3})
        self.assertEqual(c.viewers_at_or_before(1), {viewer3})

        c.detach_all()
        self.assertEqual(c.viewers_at_or_before(4), ())

        c.change(viewer1, 4)
        self.assertFalse(viewer1.detached)
        self.assertEqual(c.viewers_at_or_before(4), {viewer1})
        self.assertEqual(c.viewers_at_or_before(3), ())


    def detach_at_min(self):
        viewer1 = MockViewer()
        viewer1.highest_visible_tid = 1
        viewer2 = MockViewer()
        viewer2.highest_visible_tid = 2

        c = self._makeOne()
        c.register(viewer1)
        c.register(viewer2)

        self.assertEqual(c.maximum_highest_visible_tid, 2)
        self.assertEqual(c.minimum_highest_visible_tid, 1)

        c.detach_viewers_at_minimum()
        self.assertTrue(viewer1.detached)
        self.assertEqual(c.maximum_highest_visible_tid, 2)
        self.assertEqual(c.minimum_highest_visible_tid, 2)
